#!/usr/bin/env python
# cardinal_pythonlib/sizeformatter.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================
"""

from typing import Union


def sizeof_fmt(num: float, suffix: str = 'B') -> str:
    """
    Formats a number of bytes in a human-readable binary format (e.g. ``2048``
    becomes ``'2 KiB'``); from https://stackoverflow.com/questions/1094841.
    """
    for unit in ('', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi'):
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


# see: https://en.wikipedia.org/wiki/Binary_prefix
SYMBOLS = {
    'customary': ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'),
    'customary_ext': ('byte', 'kilo', 'mega', 'giga', 'tera', 'peta', 'exa',
                      'zetta', 'iotta'),
    'iec': ('Bi', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'),
    'iec_ext': ('byte', 'kibi', 'mebi', 'gibi', 'tebi', 'pebi', 'exbi',
                'zebi', 'yobi'),
}


def bytes2human(n: Union[int, float],
                format: str = '%(value).1f %(symbol)s',
                symbols: str = 'customary') -> str:
    """
    Converts a number of bytes into a human-readable format.
    From https://code.activestate.com/recipes/578019-bytes-to-human-human-to-bytes-converter/.

    Args:
        n: number of bytes
        format: a format specification string
        symbols: can be one of ``"customary"``, ``"customary_ext"``, ``"iec"``
            or ``"iec_ext"``; see https://goo.gl/kTQMs

    Returns:
        the formatted number

    Examples:

      >>> bytes2human(0)
      '0.0 B'
      >>> bytes2human(0.9)
      '0.0 B'
      >>> bytes2human(1)
      '1.0 B'
      >>> bytes2human(1.9)
      '1.0 B'
      >>> bytes2human(1024)
      '1.0 K'
      >>> bytes2human(1048576)
      '1.0 M'
      >>> bytes2human(1099511627776127398123789121)
      '909.5 Y'

      >>> bytes2human(9856, symbols="customary")
      '9.6 K'
      >>> bytes2human(9856, symbols="customary_ext")
      '9.6 kilo'
      >>> bytes2human(9856, symbols="iec")
      '9.6 Ki'
      >>> bytes2human(9856, symbols="iec_ext")
      '9.6 kibi'

      >>> bytes2human(10000, "%(value).1f %(symbol)s/sec")
      '9.8 K/sec'

      >>> # precision can be adjusted by playing with %f operator
      >>> bytes2human(10000, format="%(value).5f %(symbol)s")
      '9.76562 K'

    """  # noqa
    n = int(n)
    if n < 0:
        raise ValueError("n < 0")
    symbols = SYMBOLS[symbols]
    prefix = {}
    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i + 1) * 10
    for symbol in reversed(symbols[1:]):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return format % locals()
    return format % dict(symbol=symbols[0], value=n)


def human2bytes(s: str) -> int:
    """
    Modified from
    https://code.activestate.com/recipes/578019-bytes-to-human-human-to-bytes-converter/.

    Attempts to guess the string format based on default symbols
    set and return the corresponding bytes as an integer.
    When unable to recognize the format, :exc:`ValueError` is raised.

      >>> human2bytes('0 B')
      0
      >>> human2bytes('1 K')
      1024
      >>> human2bytes('1 M')
      1048576
      >>> human2bytes('1 Gi')
      1073741824
      >>> human2bytes('1 tera')
      1099511627776

      >>> human2bytes('0.5kilo')
      512
      >>> human2bytes('0.1  byte')
      0
      >>> human2bytes('1 k')  # k is an alias for K
      1024
      >>> human2bytes('12 foo')
      Traceback (most recent call last):
          ...
      ValueError: can't interpret '12 foo'
    """  # noqa
    if not s:
        raise ValueError(f"Can't interpret {s!r} as integer")
    try:
        return int(s)
    except ValueError:
        pass
    init = s
    num = ""
    while s and s[0:1].isdigit() or s[0:1] == '.':
        num += s[0]
        s = s[1:]
    num = float(num)
    letter = s.strip()
    for name, sset in SYMBOLS.items():
        if letter in sset:
            break
    else:
        if letter == 'k':
            # treat 'k' as an alias for 'K' as per https://en.wikipedia.org/wiki/Binary_prefix  # noqa
            sset = SYMBOLS['customary']
            letter = letter.upper()
        else:
            raise ValueError("can't interpret %r" % init)
    prefix = {sset[0]: 1}
    for i, s in enumerate(sset[1:]):
        prefix[s] = 1 << (i + 1) * 10
    return int(num * prefix[letter])



#!/usr/bin/env python
# cardinal_pythonlib/tsv.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Trivial functions to make tab-separated value (TSV) files.**

"""

from typing import Any, Dict, List

from cardinal_pythonlib.lists import chunks
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.text import unescape_tabs_newlines

log = get_brace_style_log_with_null_handler(__name__)


def tsv_escape(x: Any) -> str:
    """
    Escape data for tab-separated value (TSV) format.
    """
    if x is None:
        return ""
    x = str(x)
    return x.replace("\t", "\\t").replace("\n", "\\n")


def make_tsv_row(values: List[Any]) -> str:
    """
    From a list of values, make a TSV line.
    """
    return "\t".join([tsv_escape(x) for x in values]) + "\n"


def dictlist_to_tsv(dictlist: List[Dict[str, Any]]) -> str:
    """
    From a consistent list of dictionaries mapping fieldnames to values,
    make a TSV file.
    """
    if not dictlist:
        return ""
    fieldnames = dictlist[0].keys()
    tsv = "\t".join([tsv_escape(f) for f in fieldnames]) + "\n"
    for d in dictlist:
        tsv += "\t".join([tsv_escape(v) for v in d.values()]) + "\n"
    return tsv


def tsv_pairs_to_dict(line: str, key_lower: bool = True) -> Dict[str, str]:
    r"""
    Converts a TSV line into sequential key/value pairs as a dictionary.

    For example,

    .. code-block:: none

        field1\tvalue1\tfield2\tvalue2

    becomes

    .. code-block:: none

        {"field1": "value1", "field2": "value2"}

    Args:
        line: the line
        key_lower: should the keys be forced to lower case?

    Sometimes we get lines that end in a tab. This is valid. Check with these:

    .. code-block:: python

        import logging
        from cardinal_pythonlib.tsv import tsv_pairs_to_dict
        logging.basicConfig(level=logging.DEBUG)
        print(tsv_pairs_to_dict("a\t1\tb\t2\tc\t3"))  # OK
        print(tsv_pairs_to_dict("a\t1\tb\t2\tc\t"))  # OK
        print(tsv_pairs_to_dict("a\t1\tb\t2\tc"))  # not OK; orphan 'c'
        print(tsv_pairs_to_dict("a\t1\tb\t2\tc\t\n"))  # OK

    Beware using :func:`rstrip` prior to a call to this function, because that
    will also strip trailing tabs.

    """
    items = line.split("\t")
    d = {}  # type: Dict[str, str]
    for chunk in chunks(items, 2):
        if len(chunk) < 2:
            log.warning("Bad chunk, not of length 2: {!r}", chunk)
            continue
        key = chunk[0]
        value = unescape_tabs_newlines(chunk[1])
        if key_lower:
            key = key.lower()
        d[key] = value
    return d



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/table_identity.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Class to refer to database tables either by name or by SQLAlchemy Table
object.**

"""

from sqlalchemy.sql.schema import MetaData, Table

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# TableIdentity
# =============================================================================

class TableIdentity(object):
    """
    Convenient way of passing around SQLAlchemy :class:`Table` objects when you
    might know either either its name or the :class:`Table` object itself.
    """
    def __init__(self,
                 tablename: str = None,
                 table: Table = None,
                 metadata: MetaData = None) -> None:
        """
        Initialize with either ``tablename`` or ``table``, not both.

        Args:
            tablename: string name of the table
            table: SQLAlchemy :class:`Table` object
            metadata: optional :class:`MetaData` object
        """
        assert table is not None or tablename, "No table information provided"
        assert not (tablename and table is not None), (
            "Specify either table or tablename, not both")
        self._table = table
        self._tablename = tablename
        self._metadata = metadata

    def __str__(self) -> str:
        return self.tablename

    def __repr__(self) -> str:
        return (
            f"TableIdentity(table={self._table!r}, "
            f"tablename={self._tablename!r}, metadata={self._metadata!r}"
        )

    @property
    def table(self) -> Table:
        """
        Returns a SQLAlchemy :class:`Table` object. This is either the
        :class:`Table` object that was used for initialization, or one that
        was constructed from the ``tablename`` plus the ``metadata``.
        """
        if self._table is not None:
            return self._table
        assert self._metadata, (
            "Must specify metadata (in constructor or via set_metadata()/"
            "set_metadata_if_none() before you can get a Table from a "
            "tablename"
        )
        for table in self._metadata.tables.values():  # type: Table
            if table.name == self._tablename:
                return table
        raise ValueError(
            f"No table named {self._tablename!r} is present in the metadata")

    @property
    def tablename(self) -> str:
        """
        Returns the string name of the table.
        """
        if self._tablename:
            return self._tablename
        return self.table.name

    def set_metadata(self, metadata: MetaData) -> None:
        """
        Sets the :class:`MetaData`.
        """
        self._metadata = metadata

    def set_metadata_if_none(self, metadata: MetaData) -> None:
        """
        Sets the :class:`MetaData` unless it was set already.
        """
        if self._metadata is None:
            self._metadata = metadata



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/insert_on_duplicate.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Add "INSERT ON DUPLICATE KEY UPDATE" functionality to SQLAlchemy for MySQL.**

- https://www.reddit.com/r/Python/comments/p5grh/sqlalchemy_whats_the_idiomatic_way_of_writing/
- https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py
  ... modified
- https://docs.sqlalchemy.org/en/rel_1_0/core/compiler.html
- https://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update
- https://dev.mysql.com/doc/refman/5.7/en/insert-on-duplicate.html

Once implemented, you can do

.. code-block:: python

    q = sqla_table.insert_on_duplicate().values(destvalues)
    session.execute(q)

**Note: superseded by SQLAlchemy v1.2:**

- https://docs.sqlalchemy.org/en/latest/changelog/migration_12.html
- https://docs.sqlalchemy.org/en/latest/dialects/mysql.html#mysql-insert-on-duplicate-key-update

"""  # noqa

import re
from typing import Any

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.expression import Insert, TableClause

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.sqlalchemy.dialect import SqlaDialectName

log = get_brace_style_log_with_null_handler(__name__)


# noinspection PyAbstractClass
class InsertOnDuplicate(Insert):
    """
    Class that derives from :class:`Insert`, so we can hook in to operations
    involving it.
    """
    pass


def insert_on_duplicate(tablename: str,
                        values: Any = None,
                        inline: bool = False,
                        **kwargs):
    """
    Command to produce an :class:`InsertOnDuplicate` object.

    Args:
        tablename: name of the table
        values: values to ``INSERT``
        inline: as per 
            https://docs.sqlalchemy.org/en/latest/core/dml.html#sqlalchemy.sql.expression.insert
        kwargs: additional parameters

    Returns:
        an :class:`InsertOnDuplicate` object

    """  # noqa
    return InsertOnDuplicate(tablename, values, inline=inline, **kwargs)


# noinspection PyPep8Naming
def monkeypatch_TableClause() -> None:
    """
    Modifies :class:`sqlalchemy.sql.expression.TableClause` to insert
    a ``insert_on_duplicate`` member that is our :func:`insert_on_duplicate`
    function as above.
    """
    log.debug("Adding 'INSERT ON DUPLICATE KEY UPDATE' support for MySQL "
              "to SQLAlchemy")
    TableClause.insert_on_duplicate = insert_on_duplicate


# noinspection PyPep8Naming
def unmonkeypatch_TableClause() -> None:
    """
    Reverses the action of :func:`monkeypatch_TableClause`.
    """
    del TableClause.insert_on_duplicate


STARTSEPS = '`'
ENDSEPS = '`'
INSERT_FIELDNAMES_REGEX = (
    r'^INSERT\sINTO\s[{startseps}]?(?P<table>\w+)[{endseps}]?\s+'
    r'\((?P<columns>[, {startseps}{endseps}\w]+)\)\s+VALUES'.format(
        startseps=STARTSEPS, endseps=ENDSEPS)
)
# http://pythex.org/ !
RE_INSERT_FIELDNAMES = re.compile(INSERT_FIELDNAMES_REGEX)


@compiles(InsertOnDuplicate, SqlaDialectName.MYSQL)
def compile_insert_on_duplicate_key_update(insert: Insert,
                                           compiler: SQLCompiler,
                                           **kw) -> str:
    """
    Hooks into the use of the :class:`InsertOnDuplicate` class
    for the MySQL dialect. Compiles the relevant SQL for an ``INSERT...
    ON DUPLICATE KEY UPDATE`` statement.

    Notes:
 
    - We can't get the fieldnames directly from ``insert`` or ``compiler``.
    - We could rewrite the innards of the visit_insert statement
      (https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py)... 
      but, like that, it will get outdated.
    - We could use a hack-in-by-hand method
      (https://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update)
      ... but a little automation would be nice.
    - So, regex to the rescue.
    - NOTE THAT COLUMNS ARE ALREADY QUOTED by this stage; no need to repeat.
    """  # noqa
    # log.critical(compiler.__dict__)
    # log.critical(compiler.dialect.__dict__)
    # log.critical(insert.__dict__)
    s = compiler.visit_insert(insert, **kw)
    # log.critical(s)
    m = RE_INSERT_FIELDNAMES.match(s)
    if m is None:
        raise ValueError("compile_insert_on_duplicate_key_update: no match")
    columns = [c.strip() for c in m.group('columns').split(",")]
    # log.critical(columns)
    updates = ", ".join(
        [f"{c} = VALUES({c})" for c in columns])
    s += f' ON DUPLICATE KEY UPDATE {updates}'
    # log.critical(s)
    return s


_TEST_CODE = '''

from sqlalchemy import Column, String, Integer, create_engine
from sqlalchemy.orm import Session
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class OrmObject(Base):
    __tablename__ = "sometable"
    id = Column(Integer, primary_key=True)
    name = Column(String)


engine = create_engine("sqlite://", echo=True)
Base.metadata.create_all(engine)

session = Session(engine)

d1 = dict(id=1, name="One")
d2 = dict(id=2, name="Two")

insert_1 = OrmObject.__table__.insert(values=d1)
insert_2 = OrmObject.__table__.insert(values=d2)
session.execute(insert_1)
session.execute(insert_2)
session.execute(insert_1)  # raises sqlalchemy.exc.IntegrityError


# ... recommended cross-platform way is SELECT then INSERT or UPDATE 
# accordingly; see
# https://groups.google.com/forum/#!topic/sqlalchemy/aQLqeHmLPQY

'''



#!/usr/bin/env python
# cardinal_pythonlib/randomness.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Random number generation.**

"""

import base64
import os
from random import random as random_random
import secrets
import string


# =============================================================================
# Creating random strings
# =============================================================================

def create_base64encoded_randomness(num_bytes: int) -> str:
    """
    Create and return ``num_bytes`` of random data.

    The result is encoded in a string with URL-safe ``base64`` encoding.

    Used (for example) to generate session tokens.

    Which generator to use? See
    https://cryptography.io/en/latest/random-numbers/.

    Do NOT use these methods:

    .. code-block:: python

        randbytes = M2Crypto.m2.rand_bytes(num_bytes) # NO!
        randbytes = Crypto.Random.get_random_bytes(num_bytes) # NO!

    Instead, do this:

    .. code-block:: python

        randbytes = os.urandom(num_bytes)  # YES
    """
    randbytes = os.urandom(num_bytes)  # YES
    return base64.urlsafe_b64encode(randbytes).decode('ascii')


def generate_random_string(length: int,
                           characters: str = None) -> str:
    """
    Generates a random string of the specified length.
    """
    characters = characters or (
        string.ascii_letters + string.digits + string.punctuation
    )
    # We use secrets.choice() rather than random.choices() as it's better
    # for security/cryptography purposes.
    return "".join(secrets.choice(characters) for _ in range(length))


# =============================================================================
# Coin flips
# =============================================================================

def coin(p: float) -> bool:
    """
    Flips a biased coin; returns ``True`` or ``False``, with the specified
    probability being that of ``True``.
    """
    # Slower code:

    # assert 0 <= p <= 1
    # r = random.random()  # range [0.0, 1.0), i.e. 0 <= r < 1
    # return r < p

    # Check edge cases:
    # - if p == 0, impossible that r < p, since r >= 0
    # - if p == 1, always true that r < p, since r < 1

    # Faster code:

    return random_random() < p


# =============================================================================
# Testing
# =============================================================================

def _test_coin() -> None:
    """
    Tests the :func:`coin` function.
    """
    probabilities = [0, 0.25, 0.5, 0.75, 1]
    n_values = [10, 1000, 1000000]
    for p in probabilities:
        for n in n_values:
            coins = [1 if coin(p) else 0 for _ in range(n)]
            s = sum(coins)
            print(f"coin: p = {p}, n = {n} -> {s} true")


if __name__ == '__main__':
    _test_coin()



#!/usr/bin/env python
# cardinal_pythonlib/rounding.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Rounding functions.**

Note the general need to use ``Decimal``, not ``float``; otherwise rounding
errors get silly, e.g. ``-150.1 + 0.05 == -150.04999999999998``.

"""

# =============================================================================
# Imports
# =============================================================================

import decimal
from decimal import Decimal
from typing import Tuple, Union
import unittest


# =============================================================================
# Rounding/truncation
# =============================================================================

def round_half_up(x: Union[float, Decimal], dp: int = 0) -> Decimal:
    """
    Rounds, with halves going up (positive).

    (That is not the same as ``ROUND_HALF_UP`` in the ``decimal`` module!)

    See also other methods, e.g.

    - https://en.wikipedia.org/wiki/Rounding
    - https://kodify.net/python/math/round-decimals/
    - https://stackoverflow.com/questions/33019698/how-to-properly-round-up-half-float-numbers-in-python

    """  # noqa
    x = Decimal(x)
    context = decimal.getcontext()
    factor = context.power(10, dp)
    y = (
        x * factor + Decimal("0.5")
    ).quantize(Decimal("1"), rounding=decimal.ROUND_FLOOR) / factor
    # print(f"round_half_up({x}, {dp}) = {y}")
    return y


def truncate(x: Union[float, Decimal], dp: int = 0) -> Decimal:
    """
    Truncates a value to a certain number of decimal places.
    """
    x = Decimal(x)
    context = decimal.getcontext()
    factor = context.power(10, dp)
    if x >= 0:
        rounding = decimal.ROUND_FLOOR
    else:
        # Negative x
        rounding = decimal.ROUND_CEILING
    y = (x * factor).quantize(Decimal("1"), rounding=rounding) / factor
    # print(f"truncate(x={x}, dp={dp}) = {y}")
    return y


# =============================================================================
# Reverse rounding/truncation
# =============================================================================

def remove_exponent_from_decimal(d: Decimal) -> Decimal:
    """
    Converts a decimal like ``5.0E+3`` to ``5000``.
    As per https://docs.python.org/3/library/decimal.html.
    """
    return d.quantize(Decimal(1)) if d == d.to_integral() else d.normalize()


def num_dp_from_decimal(x: Decimal, with_negative_dp: bool = False) -> int:
    """
    Return the number of decimal places used by a ``Decimal``.

    By default, this is what you'd expect; e.g. ``123.45`` has 2 dp, and
    ``120`` has 0 dp. But if you set ``with_negative_dp`` to ``True``, then you
    if you pass in ``200`` you will get the answer ``-2``.

    Beware using ``str()``; Decimals can look like ``1E+2`` rather than
    ``100``.
    """
    components = str(remove_exponent_from_decimal(x)).split(".")
    if len(components) == 1:
        # No component after the decimal point.
        if with_negative_dp:
            before_period = components[0]
            num_zeros = len(before_period) - len(before_period.rstrip("0"))
            return -num_zeros
        else:
            return 0
    return len(components[1])


def range_roundable_up_to(y: Union[int, float, Decimal],
                          dp: int = 0,
                          with_description: bool = False) \
        -> Union[Tuple[Decimal, Decimal], Tuple[Decimal, Decimal, str]]:
    """
    Suppose some value ``x`` was rounded to ``y`` with ``dp`` decimal places,
    using the "round half up" rounding method (see implementation in
    :func:`round_half_up`).

    Given ``y`` and ``dp``, this function finds the range ``[a, b)``, such that
    ``a <= b``, within which ``x`` must have lain. The tuple returned is ``a,
    b``.

    If ``with_description`` is true, the tuple returned is ``a, b,
    range_description``.

    There are a large variety of rounding methods; see
    https://en.wikipedia.org/wiki/Rounding.
    Watch out -- Python's :func:`round` and Numpy's :func:`np.around` don't do
    that. See
    https://stackoverflow.com/questions/33019698/how-to-properly-round-up-half-float-numbers-in-python.

    Note that ``dp`` can be negative, as in other Python functions.

    """  # noqa
    y = Decimal(y)
    assert num_dp_from_decimal(y, with_negative_dp=True) <= dp, (
        f"Number {y} is not rounded to {dp} dp as claimed; it has "
        f"{num_dp_from_decimal(y, with_negative_dp=True)} dp"
    )
    half = Decimal("0.5") * decimal.getcontext().power(10, -dp)
    a = y - half
    b = y + half
    # print(f"range_roundable_up_to(y={y}, dp={dp}): half={half}, a={a}, b={b}")  # noqa
    if with_description:
        description = f"[{a}, {b})"
        return a, b, description
    else:
        return a, b


def range_truncatable_to(y: Union[int, float, Decimal],
                         dp: int = 0,
                         with_description: bool = False) \
        -> Union[Tuple[Decimal, Decimal], Tuple[Decimal, Decimal, str]]:
    """
    Some value ``x`` was truncated to ``y`` with ``dp`` decimal places, as per
    the implementation in :func:`truncate`. Return the range within which ``x``
    must have lain.

    The tuple returned is ``a, b``, such that ``a <= b``.

    If ``with_description`` is true, the tuple returned is ``a, b,
    range_description``.

    If ``y`` is positive, the range returned is ``[a, b)``.
    If ``y`` is negative, the range returned is ``(a, b]``.

    Note that ``dp`` can be negative, as in other Python functions.
    """
    y = Decimal(y)
    assert num_dp_from_decimal(y, with_negative_dp=True) <= dp, (
        f"Number {y} is not truncated to {dp} dp as claimed; it has "
        f"{num_dp_from_decimal(y, with_negative_dp=True)} dp"
    )
    one = decimal.getcontext().power(10, -dp)
    if y >= 0:
        a = y  # inclusive
        b = y + one  # exclusive
    else:
        # Negative y
        a = y - one  # exclusive
        b = y  # inclusive
    if with_description:
        description = f"[{a}, {b})" if y >= 0 else f"({a}, {b}]"
        return a, b, description
    else:
        return a, b


# =============================================================================
# Self-testing
# =============================================================================

def validate_range_roundable_up_to(x: float, dp: int,
                                   epsilon: Decimal = Decimal("1e-9")) -> None:
    """
    Checks assumptions for :func:`range_roundable_up_to`.

    Args:
        x: number to be rounded
        dp: number of decimal places
        epsilon: small value to check boundaries
    """
    y = round_half_up(x, dp)
    a, b, description = range_roundable_up_to(y, dp, with_description=True)
    print(
        f"validate_range_roundable_up_to: (1) x={x}, dp={dp}; "
        f"(2) y = round_half_up(x, dp) = {y}; "
        f"(3) range_roundable_up_to(y, dp) = {description}"
    )
    assert a <= x < b
    assert round_half_up(a, dp) == y, (
        f"round_half_up({a}, {dp}) is {round_half_up(a, dp)} "
        f"but should be {y}"
    )
    assert b - epsilon < b, (
        f"Use a bigger value of epsilon; currently {epsilon}"
    )
    assert round_half_up(b - epsilon, dp) == y, (
        f"round_half_up({b - epsilon}, {dp}) is "
        f"{round_half_up(b - epsilon, dp)} but should be {y}"
    )
    assert round_half_up(b, dp) > y, (
        f"round_half_up({b}, {dp}) is {round_half_up(b, dp)} "
        f"but should be >{y}"
    )


def validate_range_truncatable_to(x: float, dp: int,
                                  epsilon: Decimal = Decimal("1e-9")) -> None:
    """
    Checks assumptions for :func:`range_roundable_up_to`.

    Args:
        x: number to be truncated
        dp: number of decimal places
        epsilon: small value to check boundaries
    """
    y = truncate(x, dp)
    a, b, description = range_truncatable_to(y, dp, with_description=True)
    print(
        f"validate_range_truncatable_to: (1) x={x}, dp={dp}; "
        f"(2) y = truncate(x, dp) = {y}; "
        f"(3) range_truncatable_to(y, dp) = {description}"
    )
    if y >= 0:
        assert a <= x < b
        assert truncate(a, dp) == y, (
            f"truncate({a}, {dp}) is {truncate(a, dp)} "
            f"but should be {y}"
        )
        assert b - epsilon < b, (
            f"Use a bigger value of epsilon; currently {epsilon}"
        )
        assert truncate(b - epsilon, dp) == y, (
            f"truncate({b - epsilon}, {dp}) is "
            f"{truncate(b - epsilon, dp)} but should be {y}"
        )
        assert truncate(b, dp) > y, (
            f"truncate({b}, {dp}) is {truncate(b, dp)} "
            f"but should be >{y}"
        )
    else:
        # Negative values
        assert a < x <= b
        assert a + epsilon > a, (
            f"Use a bigger value of epsilon; currently {epsilon}"
        )
        assert truncate(a, dp) < y, (
            f"truncate({a}, {dp}) is {truncate(a, dp)} "
            f"but should be <{y}"
        )
        assert truncate(a + epsilon, dp) == y, (
            f"truncate({a + epsilon}, {dp}) is "
            f"{truncate(a + epsilon, dp)} but should be {y}"
        )
        assert truncate(b, dp) == y, (
            f"truncate({b}, {dp}) is {truncate(b, dp)} "
            f"but should be {y}"
        )


class TestRoundingAndReversal(unittest.TestCase):
    EPSILON = Decimal("1e-9")

    def test_round_half_up(self) -> None:
        assert round_half_up(Decimal("-123.51"), 0) == Decimal("-124.0")
        assert round_half_up(Decimal("-123.5"), 0) == Decimal("-123.0")
        assert round_half_up(Decimal("-123.49"), 0) == Decimal("-123.0")

        assert round_half_up(Decimal("-123.456"), -1) == Decimal("-120.0")
        assert round_half_up(Decimal("-123.456"), 0) == Decimal("-123.0")
        assert round_half_up(Decimal("-123.456"), 1) == Decimal("-123.5")

        assert round_half_up(Decimal("-0.51"), 0) == Decimal("-1.0")
        assert round_half_up(Decimal("-0.5"), 0) == Decimal("0.0")
        assert round_half_up(Decimal("-0.49"), 0) == Decimal("0.0")
        assert round_half_up(Decimal("0.49"), 0) == Decimal("0.0")
        assert round_half_up(Decimal("0.49"), 0) == Decimal("0.0")
        assert round_half_up(Decimal("0.5"), 0) == Decimal("1.0")

        assert round_half_up(Decimal("123.456"), -1) == Decimal("120.0")
        assert round_half_up(Decimal("123.456"), 0) == Decimal("123.0")
        assert round_half_up(Decimal("123.456"), 1) == Decimal("123.5")

    def test_truncate(self) -> None:

        assert truncate(Decimal("-123.456"), -1) == Decimal("-120.0")
        assert truncate(Decimal("-123.456"), 0) == Decimal("-123.0")
        assert truncate(Decimal("-123.456"), 1) == Decimal("-123.4")

        assert truncate(Decimal("-0.51"), 0) == Decimal("0.0")
        assert truncate(Decimal("-0.5"), 0) == Decimal("0.0")
        assert truncate(Decimal("-0.49"), 0) == Decimal("0.0")
        assert truncate(Decimal("0.49"), 0) == Decimal("0.0")
        assert truncate(Decimal("0.49"), 0) == Decimal("0.0")
        assert truncate(Decimal("0.5"), 0) == Decimal("0.0")

        assert truncate(Decimal("123.456"), -1) == Decimal("120.0")
        assert truncate(Decimal("123.456"), 0) == Decimal("123.0")
        assert truncate(Decimal("123.456"), 1) == Decimal("123.4")

    def test_range_roundable_up_to(self) -> None:

        assert range_roundable_up_to(Decimal("200"), -2) == \
               (Decimal("150.0"), Decimal("250.0"))
        assert range_roundable_up_to(Decimal("200"), -1) == \
               (Decimal("195.0"), Decimal("205.0"))
        assert range_roundable_up_to(Decimal("200"), 0) == \
               (Decimal("199.5"), Decimal("200.5"))
        assert range_roundable_up_to(Decimal("200"), 1) == \
               (Decimal("199.95"), Decimal("200.05"))
        assert range_roundable_up_to(Decimal("200"), 2) == \
               (Decimal("199.995"), Decimal("200.005"))

        assert range_roundable_up_to(Decimal("-1"), 0) == \
               (Decimal("-1.5"), Decimal("-0.5"))

        # range_roundable_up_to(0.5, 0)  # bad input (not correctly rounded); would assert  # noqa

        for x in [Decimal("100.332"), Decimal("-150.12")]:
            for dp in [-2, -1, 0, 1, 2]:
                validate_range_roundable_up_to(x, dp, self.EPSILON)

    def test_range_truncatable_to(self) -> None:

        assert range_truncatable_to(Decimal("200"), -2) == \
               (Decimal("200"), Decimal("300"))
        assert range_truncatable_to(Decimal("200"), -1) == \
               (Decimal("200"), Decimal("210"))
        assert range_truncatable_to(Decimal("200"), 0) == \
               (Decimal("200"), Decimal("201"))
        assert range_truncatable_to(Decimal("200"), 1) == \
               (Decimal("200"), Decimal("200.1"))
        assert range_truncatable_to(Decimal("200"), 2) == \
               (Decimal("200"), Decimal("200.01"))

        assert range_truncatable_to(Decimal("-1"), 0) == \
               (Decimal("-2"), Decimal("-1"))
        assert range_truncatable_to(Decimal("-1"), 1) == \
               (Decimal("-1.1"), Decimal("-1"))
        assert range_truncatable_to(Decimal("-1"), 2) == \
               (Decimal("-1.01"), Decimal("-1"))

        # range_truncatable_to(0.5, 0)  # bad input (not correctly rounded); would assert  # noqa

        for x in [Decimal("100.332"), Decimal("-150.12")]:
            for dp in [-2, -1, 0, 1, 2]:
                validate_range_truncatable_to(x, dp, self.EPSILON)


# =============================================================================
# Command-line entry point
# =============================================================================

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# cardinal_pythonlib/probability.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Miscellaneous probability functions.**

"""

import math
from math import (
    exp as math_exp,
    log as math_ln,
    log10 as math_log10,
)

# For speed:
INFINITY = math.inf
MINUS_INFINITY = -math.inf


# =============================================================================
# Basic maths
# =============================================================================

def ln(x: float) -> float:
    """
    Version of :func:`math.log` that treats log(0) as ``-inf``, rather than
    crashing with ``ValueError: math domain error``.

    Args:
        x: parameter

    Returns:
        float: ln(x), the natural logarithm of x

    See
    https://stackoverflow.com/questions/42980201/logarithm-of-zero-in-python.

    For speed, use ``from math import log as math_log``, etc.:

    .. code-block: python

        import math
        from math import log as math_log
        import timeit
        INFINITY = math.inf
        MINUS_INFINITY = -math.inf

        timeit.timeit('(math.log(x) for x in range(1, 100))', number=10000)
        # 0.00756470113992691
        timeit.timeit('(math_log(x) for x in range(1, 100))', number=10000)
        # 0.004151405766606331

        timeit.timeit('(-math.inf for _ in range(1, 100))', number=10000)
        # 0.009433002211153507
        timeit.timeit('(MINUS_INFINITY for _ in range(1, 100))', number=10000)
        # 0.004293474368751049

    """
    return math_ln(x) if x != 0 else MINUS_INFINITY


def log10(x: float) -> float:
    """
    Version of :func:`math.log10` that treats log(0) as ``-inf``, rather than
    crashing with ``ValueError: math domain error``.

    Args:
        x: parameter

    Returns:
        float: log10(x), the logarithm to base 10 of x

    See
    https://stackoverflow.com/questions/42980201/logarithm-of-zero-in-python.
    """
    return math_log10(x) if x != 0 else MINUS_INFINITY


# =============================================================================
# Basic probability
# =============================================================================

def odds_from_probability(p: float) -> float:
    r"""
    Returns odds, given a probability.

    .. math::

        o = \frac{p}{1 - p}

    Args:
        p: probability

    Returns:
        float: odds
    """
    return p / (1 - p)


def probability_from_odds(odds: float) -> float:
    r"""
    Returns a probability, given odds.

    .. math::

        o &= \frac{p}{1 - p}  \\
        o (1 - p) &= p  \\
        o - op &= p  \\
        o &= p + op  \\
        o &= p(1 + o)  \\
        p &= \frac{o}{1 + o}

    Args:
        odds: odds

    Returns:
        float: probability
    """
    return odds / (1 + odds)


def probability_from_log_prob(log_p: float) -> float:
    """
    Returns a probability from a log probability.

    Args:
        log_p: ln(p); natural log of p

    Returns:
        float: p
    """
    return math_exp(log_p)


def probability_from_log_odds(log_odds: float) -> float:
    """
    Returns a probability from a log odds.

    Args:
        log_odds: ln(o); natural log of o

    Returns:
        float: p
    """
    return probability_from_odds(math_exp(log_odds))


def log_probability_from_log_odds(log_odds: float) -> float:
    r"""
    No obvious quick form for this:

    .. math::

        o &= \frac{p}{1 - p}  \\

        \log o &= \log p - \log (1 - p)  \\

    Args:
        log_odds: ln(o); natural log of o

    Returns:
        float: ln(p)

    """
    return ln(probability_from_log_odds(log_odds))


def log_odds_from_probability(p: float) -> float:
    """
    Returns log ods from a probability.

    Args:
        p: probability

    Returns:
        float: ln(odds)

    See

    - https://wiki.lesswrong.com/wiki/Odds_ratio
    - https://wiki.lesswrong.com/wiki/Log_odds
    """
    return ln(odds_from_probability(p))


def log_odds_from_1_in_n(n: float) -> float:
    r"""
    If the chance of something occurring are 1 in n, then its probability is
    :math:`1/n`, and its odds are :math:`\frac{1}{n - 1}`. This function
    returns the log of those odds.

    Args:
        n: ``n``, as above

    Returns:
        float: :math:`\ln \frac{1}{n - 1} = \ln 1 - \ln (n - 1) = -\ln(n - 1)`
    """
    return -ln(n - 1)


# =============================================================================
# Bayesian update rules
# =============================================================================

def bayes_posterior(prior: float, likelihood: float,
                    marginal_likelihood: float) -> float:
    r"""
    Returns P(H | D), the posterior probability of hypothesis H, given data D.

    Args:
        prior:
            P(H), the prior belief in H
        likelihood:
            P(D | H), the probability of observing D given H
        marginal_likelihood:
            P(D), the probability of observing D; also called evidence or model
            evidence.

    Returns:
        float: P(H | D), the probability of H given D.

    Bayes' rule:

    .. math::

        P(A | B) P(B) = P(B | A) P(A) = P(A \land B)

        P(H | D) = \frac{ P(D | H) \cdot P(H) }{ P(D) }

        \text{posterior} =
            \frac
            { \text{likelihood} \cdot \text{prior} }
            { \text{marginal likelihood} }

    """
    return likelihood * prior / marginal_likelihood


def log_bayes_posterior(log_prior: float, log_likelihood: float,
                        log_marginal_likelihood: float) -> float:
    r"""
    Exactly equivalent to :func:`bayes_posterior`, but using log probability.

    Args:
        log_prior: :math:`\log P(H)`
        log_likelihood: :math:`\log P(D | H)`
        log_marginal_likelihood: :math:`\log P(D)`

    Returns:
        float: :math:`\log P(H | D)`
    """
    return log_likelihood + log_prior - log_marginal_likelihood


def posterior_odds(prior_odds: float,
                   likelihood_ratio: float) -> float:
    r"""
    Returns the posterior odds for a hypothesis, given the prior odds and the
    likelihood ratio.

    .. math::

        P(A | B) &= P(B | A) \frac{ P(A) }{ P(B) }  \\

        P(\neg A | B) &= P(B | \neg A) \frac{ P(\neg A) }{ P(B) }  \\

        \frac{ P(A | B) }{ P(\neg A | B) } &=
            \frac{ P(B | A) }{ P(B | \neg A) }
            \frac{ P(A) }{ P(\neg A) }  \\

        \text{posterior odds} &= \text{likelihood ratio} \cdot
                                 \text{prior odds}

    Args:
        prior_odds:
            prior odds, :math:`\frac{ P(A) }{ P(\neg A) }`
        likelihood_ratio:
            likelihood ratio, :math:`\frac{ P(B | A) }{ P(B | \neg A) }`

    Returns:
        float: posterior odds, :math:`\frac{ P(A | B) }{ P(\neg A | B) }`

    e.g. https://wiki.lesswrong.com/wiki/Odds_ratio

    """
    return likelihood_ratio * prior_odds


def log_posterior_odds(log_prior_odds: float,
                       log_likelihood_ratio: float) -> float:
    r"""
    Exactly as for :func:`posterior_odds`, but with log odds.
    """
    return log_likelihood_ratio + log_prior_odds


def log_likelihood_ratio_from_p(p_d_given_h: float,
                                p_d_given_not_h: float) -> float:
    r"""
    Returns

    .. math::

        \ln \frac{ P(D | H) }{ P(D | \neg H) } =
            \ln P(D | H) - \ln P(D | \neg H)

    Args:
        p_d_given_h: :math:`P(D | H)`
        p_d_given_not_h: :math:`P(D | \neg H)`

    Returns:
        float: log likelihood ratio, as above

    """  # noqa
    return ln(p_d_given_h) - ln(p_d_given_not_h)


def log_posterior_odds_from_pdh_pdnh(log_prior_odds: float,
                                     p_d_given_h: float,
                                     p_d_given_not_h: float) -> float:
    r"""
    Calculates posterior odds.

    Args:
        log_prior_odds:
            log prior odds of H, :math:`ln(\frac{ P(H) }{ P(\neg H) })`
        p_d_given_h:
            :math:`P(D | H)`
        p_d_given_not_h:
            :math:`P(D | \neg H)`

    Returns:
        float:
            posterior odds of H, :math:`ln(\frac{ P(H | D) }{ P(\neg H | D) })`

    """
    return log_posterior_odds(
        log_prior_odds=log_prior_odds,
        log_likelihood_ratio=log_likelihood_ratio_from_p(
            p_d_given_h=p_d_given_h,
            p_d_given_not_h=p_d_given_not_h
        )
    )


def log_posterior_odds_from_bool_d_pdh_pdnh(log_prior_odds: float,
                                            d: bool,
                                            p_d_given_h: float,
                                            p_d_given_not_h: float) -> float:
    r"""
    Calculates posterior odds.

    Args:
        log_prior_odds:
            log prior odds of H, :math:`ln(\frac{ P(H) }{ P(\neg H) })`
        d:
            whether D is true or not
        p_d_given_h:
            :math:`P(D | H)`
        p_d_given_not_h:
            :math:`P(D | \neg H)`

    Returns:
        float:
            posterior odds of H, :math:`ln(\frac{ P(H | D) }{ P(\neg H | D) })`

    """
    if d:
        # D is true
        log_lr = log_likelihood_ratio_from_p(
            p_d_given_h=p_d_given_h,
            p_d_given_not_h=p_d_given_not_h
        )
    else:
        # not-D is true
        p_not_d_given_h = 1 - p_d_given_h
        p_not_d_given_not_h = 1 - p_d_given_not_h
        log_lr = log_likelihood_ratio_from_p(
            p_d_given_h=p_not_d_given_h,
            p_d_given_not_h=p_not_d_given_not_h
        )
    return log_posterior_odds(
        log_prior_odds=log_prior_odds,
        log_likelihood_ratio=log_lr
    )



#!/usr/bin/env python3
# cardinal_pythonlib/openxml/pause_process_by_disk_space.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Pauses and resumes a process by disk space; LINUX ONLY.**

"""

from argparse import ArgumentParser
import logging
import shutil
import subprocess
import sys
from time import sleep
from typing import NoReturn

from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)
from cardinal_pythonlib.sizeformatter import human2bytes, sizeof_fmt

log = BraceStyleAdapter(logging.getLogger(__name__))


def is_running(process_id: int) -> bool:
    """
    Uses the Unix ``ps`` program to see if a process is running.
    """
    pstr = str(process_id)
    encoding = sys.getdefaultencoding()
    s = subprocess.Popen(["ps", "-p", pstr], stdout=subprocess.PIPE)
    for line in s.stdout:
        strline = line.decode(encoding)
        if pstr in strline:
            return True
    return False


def main() -> NoReturn:
    """
    Command-line handler for the ``pause_process_by_disk_space`` tool.
    Use the ``--help`` option for help.
    """
    parser = ArgumentParser(
        description="Pauses and resumes a process by disk space; LINUX ONLY."
    )
    parser.add_argument(
        "process_id", type=int,
        help="Process ID."
    )
    parser.add_argument(
        "--path", required=True,
        help="Path to check free space for (e.g. '/')"
    )
    parser.add_argument(
        "--pause_when_free_below", type=str, required=True,
        help="Pause process when free disk space below this value (in bytes "
             "or as e.g. '50G')"
    )
    parser.add_argument(
        "--resume_when_free_above", type=str, required=True,
        help="Resume process when free disk space above this value (in bytes "
             "or as e.g. '70G')"
    )
    parser.add_argument(
        "--check_every", type=int, required=True,
        help="Check every n seconds (where this is n)"
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Verbose output"
    )
    args = parser.parse_args()
    main_only_quicksetup_rootlogger(
        level=logging.DEBUG if args.verbose else logging.INFO)

    minimum = human2bytes(args.pause_when_free_below)
    maximum = human2bytes(args.resume_when_free_above)
    path = args.path
    process_id = args.process_id
    period = args.check_every
    pause_args = ["kill", "-STOP", str(process_id)]
    resume_args = ["kill", "-CONT", str(process_id)]

    assert minimum < maximum, "Minimum must be less than maximum"

    log.info(
        f"Starting: controlling process {process_id}; "
        f"checking disk space every {period} s; "
        f"will pause when free space on {path} "
        f"is less than {sizeof_fmt(minimum)} and "
        f"resume when free space is at least {sizeof_fmt(maximum)}; "
        f"pause command will be {pause_args}; "
        f"resume command will be {resume_args}."
    )
    log.debug("Presuming that the process is RUNNING to begin with.")

    paused = False
    while True:
        if not is_running(process_id):
            log.info("Process {} is no longer running", process_id)
            sys.exit(0)
        space = shutil.disk_usage(path).free
        log.debug("Disk space on {} is {}", path, sizeof_fmt(space))
        if space < minimum and not paused:
            log.info("Disk space down to {}: pausing process {}",
                     sizeof_fmt(space), process_id)
            subprocess.check_call(pause_args)
            paused = True
        elif space >= maximum and paused:
            log.info("Disk space up to {}: resuming process {}",
                     sizeof_fmt(space), process_id)
            subprocess.check_call(resume_args)
            paused = False
        log.debug("Sleeping for {} seconds...", period)
        sleep(period)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/rnc_text.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Low-quality functions relating to textfile results storage/analysis.**

"""

import csv
import datetime
import logging
from typing import Any, Dict, Iterable, List, Optional, Sequence, TextIO, Tuple

from cardinal_pythonlib.logs import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__name__))


def produce_csv_output(filehandle: TextIO,
                       fields: Sequence[str],
                       values: Iterable[str]) -> None:
    """
    Produce CSV output, without using ``csv.writer``, so the log can be used
    for lots of things.

    - ... eh? What was I talking about?
    - POOR; DEPRECATED.

    Args:
        filehandle: file to write to
        fields: field names
        values: values
    """
    output_csv(filehandle, fields)
    for row in values:
        output_csv(filehandle, row)


def output_csv(filehandle: TextIO, values: Iterable[str]) -> None:
    """
    Write a line of CSV. POOR; does not escape things properly. DEPRECATED.

    Args:
        filehandle: file to write to
        values: values
    """
    line = ",".join(values)
    filehandle.write(line + "\n")


def get_what_follows_raw(s: str,
                         prefix: str,
                         onlyatstart: bool = True,
                         stripwhitespace: bool = True) -> Tuple[bool, str]:
    """
    Find the part of ``s`` that is after ``prefix``.

    Args:
        s: string to analyse
        prefix: prefix to find
        onlyatstart: only accept the prefix if it is right at the start of
            ``s``
        stripwhitespace: remove whitespace from the result

    Returns:
        tuple: ``(found, result)``

    """
    prefixstart = s.find(prefix)
    if ((prefixstart == 0 and onlyatstart) or
            (prefixstart != -1 and not onlyatstart)):
        # substring found
        resultstart = prefixstart + len(prefix)
        result = s[resultstart:]
        if stripwhitespace:
            result = result.strip()
        return True, result
    return False, ""


def get_what_follows(strings: Sequence[str],
                     prefix: str,
                     onlyatstart: bool = True,
                     stripwhitespace: bool = True,
                     precedingline: str = "") -> str:
    """
    Find a string in ``strings`` that begins with ``prefix``; return the part
    that's after ``prefix``. Optionally, require that the preceding string
    (line) is ``precedingline``.

    Args:
        strings: strings to analyse
        prefix: prefix to find
        onlyatstart: only accept the prefix if it is right at the start of
            ``s``
        stripwhitespace: remove whitespace from the result
        precedingline: if truthy, require that the preceding line be as
            specified here

    Returns:
        the line fragment

    """
    if not precedingline:
        for s in strings:
            (found, result) = get_what_follows_raw(s, prefix, onlyatstart,
                                                   stripwhitespace)
            if found:
                return result
        return ""
    else:
        for i in range(1, len(strings)):  # i indexes the second of a pair
            if strings[i-1].find(precedingline) == 0:
                # ... if found at the start
                (found, result) = get_what_follows_raw(strings[i], prefix,
                                                       onlyatstart,
                                                       stripwhitespace)
                if found:
                    return result
        return ""


def get_string(strings: Sequence[str],
               prefix: str,
               ignoreleadingcolon: bool = False,
               precedingline: str = "") -> Optional[str]:
    """
    Find a string as per :func:`get_what_follows`.

    Args:
        strings: see :func:`get_what_follows`
        prefix: see :func:`get_what_follows`
        ignoreleadingcolon: if ``True``, restrict the result to what comes
            after its first colon (and whitespace-strip that)
        precedingline: see :func:`get_what_follows`

    Returns:
        the line fragment

    """
    s = get_what_follows(strings, prefix, precedingline=precedingline)
    if ignoreleadingcolon:
        f = s.find(":")
        if f != -1:
            s = s[f+1:].strip()
    if len(s) == 0:
        return None
    return s


def get_string_relative(strings: Sequence[str],
                        prefix1: str,
                        delta: int,
                        prefix2: str,
                        ignoreleadingcolon: bool = False,
                        stripwhitespace: bool = True) -> Optional[str]:
    """
    Finds a line (string) in ``strings`` beginning with ``prefix1``. Moves
    ``delta`` lines (strings) further. Returns the end of the line that
    begins with ``prefix2``, if found.

    Args:
        strings: as above
        prefix1: as above
        delta: as above
        prefix2: as above
        ignoreleadingcolon: restrict the result to the part after its first
            colon?
        stripwhitespace: strip whitespace from the start/end of the result?

    Returns:
        the line fragment
    """
    for firstline in range(0, len(strings)):
        if strings[firstline].find(prefix1) == 0:  # if found...
            secondline = firstline + delta
            if secondline < 0 or secondline >= len(strings):
                continue
            if strings[secondline].find(prefix2) == 0:
                s = strings[secondline][len(prefix2):]
                if stripwhitespace:
                    s = s.strip()
                if ignoreleadingcolon:
                    f = s.find(":")
                    if f != -1:
                        s = s[f+1:].strip()
                    if stripwhitespace:
                        s = s.strip()
                if len(s) == 0:
                    return None
                return s
    return None


def get_int(strings: Sequence[str],
            prefix: str,
            ignoreleadingcolon: bool = False,
            precedingline: str = "") -> Optional[int]:
    """
    Fetches an integer parameter via :func:`get_string`.
    """
    return get_int_raw(get_string(strings, prefix,
                                  ignoreleadingcolon=ignoreleadingcolon,
                                  precedingline=precedingline))


def get_float(strings: Sequence[str],
              prefix: str,
              ignoreleadingcolon: bool = False,
              precedingline: str = "") -> Optional[float]:
    """
    Fetches a float parameter via :func:`get_string`.
    """
    return get_float_raw(get_string(strings, prefix,
                                    ignoreleadingcolon=ignoreleadingcolon,
                                    precedingline=precedingline))


def get_int_raw(s: str) -> Optional[int]:
    """
    Converts its input to an int.

    Args:
        s: string

    Returns:
        ``int(s)``, or ``None`` if ``s`` is ``None``

    Raises:
        ValueError: if it's a bad string

    """
    if s is None:
        return None
    return int(s)


def get_bool_raw(s: str) -> Optional[bool]:
    """
    Maps ``"Y"``, ``"y"`` to ``True`` and ``"N"``, ``"n"`` to ``False``.
    """
    if s == "Y" or s == "y":
        return True
    elif s == "N" or s == "n":
        return False
    return None


def get_float_raw(s: str) -> Optional[float]:
    """
    Converts its input to a float.

    Args:
        s: string

    Returns:
        ``int(s)``, or ``None`` if ``s`` is ``None``

    Raises:
        ValueError: if it's a bad string

    """
    if s is None:
        return None
    return float(s)


def get_bool(strings: Sequence[str],
             prefix: str,
             ignoreleadingcolon: bool = False,
             precedingline: str = "") -> Optional[bool]:
    """
    Fetches a boolean parameter via :func:`get_string`.
    """
    return get_bool_raw(get_string(strings, prefix,
                                   ignoreleadingcolon=ignoreleadingcolon,
                                   precedingline=precedingline))


def get_bool_relative(strings: Sequence[str],
                      prefix1: str,
                      delta: int,
                      prefix2: str,
                      ignoreleadingcolon: bool = False) -> Optional[bool]:
    """
    Fetches a boolean parameter via :func:`get_string_relative`.
    """
    return get_bool_raw(get_string_relative(
        strings, prefix1, delta, prefix2,
        ignoreleadingcolon=ignoreleadingcolon))


def get_float_relative(strings: Sequence[str],
                       prefix1: str,
                       delta: int,
                       prefix2: str,
                       ignoreleadingcolon: bool = False) -> Optional[float]:
    """
    Fetches a float parameter via :func:`get_string_relative`.
    """
    return get_float_raw(get_string_relative(
        strings, prefix1, delta, prefix2,
        ignoreleadingcolon=ignoreleadingcolon))


def get_int_relative(strings: Sequence[str],
                     prefix1: str,
                     delta: int,
                     prefix2: str,
                     ignoreleadingcolon: bool = False) -> Optional[int]:
    """
    Fetches an int parameter via :func:`get_string_relative`.
    """
    return get_int_raw(get_string_relative(
        strings, prefix1, delta, prefix2,
        ignoreleadingcolon=ignoreleadingcolon))


def get_datetime(strings: Sequence[str],
                 prefix: str,
                 datetime_format_string: str,
                 ignoreleadingcolon: bool = False,
                 precedingline: str = "") -> Optional[datetime.datetime]:
    """
    Fetches a ``datetime.datetime`` parameter via :func:`get_string`.
    """
    x = get_string(strings, prefix, ignoreleadingcolon=ignoreleadingcolon,
                   precedingline=precedingline)
    if len(x) == 0:
        return None
    # For the format strings you can pass to datetime.datetime.strptime, see
    # http://docs.python.org/library/datetime.html
    # A typical one is "%d-%b-%Y (%H:%M:%S)"
    d = datetime.datetime.strptime(x, datetime_format_string)
    return d


def find_line_beginning(strings: Sequence[str],
                        linestart: Optional[str]) -> int:
    """
    Finds the index of the line in ``strings`` that begins with ``linestart``,
    or ``-1`` if none is found.

    If ``linestart is None``, match an empty line.
    """
    if linestart is None:  # match an empty line
        for i in range(len(strings)):
            if is_empty_string(strings[i]):
                return i
        return -1
    for i in range(len(strings)):
        if strings[i].find(linestart) == 0:
            return i
    return -1


def find_line_containing(strings: Sequence[str], contents: str) -> int:
    """
    Finds the index of the line in ``strings`` that contains ``contents``,
    or ``-1`` if none is found.
    """
    for i in range(len(strings)):
        if strings[i].find(contents) != -1:
            return i
    return -1


def get_lines_from_to(strings: List[str],
                      firstlinestart: str,
                      list_of_lastline_starts: Iterable[Optional[str]]) \
        -> List[str]:
    """
    Takes a list of ``strings``. Returns a list of strings FROM
    ``firstlinestart`` (inclusive) TO the first of ``list_of_lastline_starts``
    (exclusive).

    To search to the end of the list, use ``list_of_lastline_starts = []``.

    To search to a blank line, use ``list_of_lastline_starts = [None]``
    """
    start_index = find_line_beginning(strings, firstlinestart)
    # log.debug("start_index: {}", start_index)
    if start_index == -1:
        return []
    end_offset = None  # itself a valid slice index
    for lls in list_of_lastline_starts:
        possible_end_offset = find_line_beginning(strings[start_index:], lls)
        # log.debug("lls {!r} -> possible_end_offset {}",
        #           lls, possible_end_offset)
        if possible_end_offset != -1:  # found one
            if end_offset is None or possible_end_offset < end_offset:
                end_offset = possible_end_offset
    end_index = None if end_offset is None else (start_index + end_offset)
    # log.debug("end_index: {}", end_index)
    return strings[start_index:end_index]


def is_empty_string(s: str) -> bool:
    """
    Is the string empty (ignoring whitespace)?
    """
    return len(s.strip()) == 0


def csv_to_list_of_fields(lines: List[str],
                          csvheader: str,
                          quotechar: str = '"') -> List[List[str]]:
    """
    Extracts data from a list of CSV lines (starting with a defined header
    line) embedded in a longer text block but ending with a blank line.

    Used for processing e.g. MonkeyCantab rescue text output.

    Args:
        lines: CSV lines
        csvheader: CSV header line
        quotechar: ``quotechar`` parameter passed to :func:`csv.reader`

    Returns:
        list (by row) of lists (by value); see example

    Test code:

    .. code-block:: python

        import logging
        from cardinal_pythonlib.rnc_text import *
        logging.basicConfig(level=logging.DEBUG)

        myheader = "field1,field2,field3"
        mycsvlines = [
            "irrelevant line",
            myheader,  # header: START
            "row1value1,row1value2,row1value3",
            "row2value1,row2value2,row2value3",
            "",  # terminating blank line: END
            "other irrelevant line",
        ]
        csv_to_list_of_fields(mycsvlines, myheader)
        # [['row1value1', 'row1value2', 'row1value3'], ['row2value1', 'row2value2', 'row2value3']]

    """  # noqa
    data = []  # type: List[List[str]]
    # an empty line marks the end of the block
    csvlines = get_lines_from_to(lines, csvheader, [None])[1:]
    # ... remove the CSV header
    reader = csv.reader(csvlines, quotechar=quotechar)
    for fields in reader:
        data.append(fields)
    return data


def csv_to_list_of_dicts(lines: List[str],
                         csvheader: str,
                         quotechar: str = '"') -> List[Dict[str, str]]:
    """
    Extracts data from a list of CSV lines (starting with a defined header
    line) embedded in a longer text block but ending with a blank line.

    Args:
        lines: CSV lines
        csvheader: CSV header line
        quotechar: ``quotechar`` parameter passed to :func:`csv.reader`

    Returns:
        list of dictionaries mapping fieldnames (from the header) to values

    """
    data = []  # type: List[Dict[str, str]]
    # an empty line marks the end of the block
    csvlines = get_lines_from_to(lines, csvheader, [None])[1:]
    # ... remove the CSV header
    headerfields = csvheader.split(",")
    reader = csv.reader(csvlines, quotechar=quotechar)
    for fields in reader:
        row = {}  # type: Dict[str, str]
        for f in range(len(headerfields)):
            row[headerfields[f]] = fields[f]
        data.append(row)
    return data


def dictlist_convert_to_string(dict_list: Iterable[Dict], key: str) -> None:
    """
    Process an iterable of dictionaries. For each dictionary ``d``, convert
    (in place) ``d[key]`` to a string form, ``str(d[key])``. If the result is a
    blank string, convert it to ``None``.
    """
    for d in dict_list:
        d[key] = str(d[key])
        if d[key] == "":
            d[key] = None


def dictlist_convert_to_datetime(dict_list: Iterable[Dict],
                                 key: str,
                                 datetime_format_string: str) -> None:
    """
    Process an iterable of dictionaries. For each dictionary ``d``, convert
    (in place) ``d[key]`` to a ``datetime.datetime`` form, using
    ``datetime_format_string`` as the format parameter to
    :func:`datetime.datetime.strptime`.
    """
    for d in dict_list:
        d[key] = datetime.datetime.strptime(d[key], datetime_format_string)


def dictlist_convert_to_int(dict_list: Iterable[Dict], key: str) -> None:
    """
    Process an iterable of dictionaries. For each dictionary ``d``, convert
    (in place) ``d[key]`` to an integer. If that fails, convert it to ``None``.
    """
    for d in dict_list:
        try:
            d[key] = int(d[key])
        except ValueError:
            d[key] = None


def dictlist_convert_to_float(dict_list: Iterable[Dict], key: str) -> None:
    """
    Process an iterable of dictionaries. For each dictionary ``d``, convert
    (in place) ``d[key]`` to a float. If that fails, convert it to ``None``.
    """
    for d in dict_list:
        try:
            d[key] = float(d[key])
        except ValueError:
            d[key] = None


def dictlist_convert_to_bool(dict_list: Iterable[Dict], key: str) -> None:
    """
    Process an iterable of dictionaries. For each dictionary ``d``, convert
    (in place) ``d[key]`` to a bool. If that fails, convert it to ``None``.
    """
    for d in dict_list:
        # d[key] = True if d[key] == "Y" else False
        d[key] = 1 if d[key] == "Y" else 0


def dictlist_replace(dict_list: Iterable[Dict], key: str, value: Any) -> None:
    """
    Process an iterable of dictionaries. For each dictionary ``d``, change
    (in place) ``d[key]`` to ``value``.
    """
    for d in dict_list:
        d[key] = value


def dictlist_wipe_key(dict_list: Iterable[Dict], key: str) -> None:
    """
    Process an iterable of dictionaries. For each dictionary ``d``, delete
    ``d[key]`` if it exists.
    """
    for d in dict_list:
        d.pop(key, None)



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/orm_inspect.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to inspect and copy SQLAlchemy ORM objects.**

"""

from typing import (Dict, Generator, List, Set, Tuple, Type, TYPE_CHECKING,
                    Union)

# noinspection PyProtectedMember
from sqlalchemy.ext.declarative.base import _get_immediate_cls_attr
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.base import class_mapper
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.schema import Column, MetaData
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.sql.visitors import VisitableType
from sqlalchemy.util import OrderedProperties

from cardinal_pythonlib.classes import gen_all_subclasses
from cardinal_pythonlib.enumlike import OrderedNamespace
from cardinal_pythonlib.dicts import reversedict
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

if TYPE_CHECKING:
    from sqlalchemy.orm.state import InstanceState
    from sqlalchemy.sql.schema import Table

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Creating ORM objects conveniently, etc.
# =============================================================================

def coltype_as_typeengine(coltype: Union[VisitableType,
                                         TypeEngine]) -> TypeEngine:
    """
    Instances of SQLAlchemy column types are subclasses of ``TypeEngine``.
    It's possible to specify column types either as such instances, or as the
    class type. This function ensures that such classes are converted to
    instances.

    To explain: you can specify columns like

    .. code-block:: python

        a = Column("a", Integer)
        b = Column("b", Integer())
        c = Column("c", String(length=50))

        isinstance(Integer, TypeEngine)  # False
        isinstance(Integer(), TypeEngine)  # True
        isinstance(String(length=50), TypeEngine)  # True

        type(Integer)  # <class 'sqlalchemy.sql.visitors.VisitableType'>
        type(Integer())  # <class 'sqlalchemy.sql.sqltypes.Integer'>
        type(String)  # <class 'sqlalchemy.sql.visitors.VisitableType'>
        type(String(length=50))  # <class 'sqlalchemy.sql.sqltypes.String'>

    This function coerces things to a :class:`TypeEngine`.
    """
    if isinstance(coltype, TypeEngine):
        return coltype
    return coltype()  # type: TypeEngine


# =============================================================================
# SqlAlchemyAttrDictMixin
# =============================================================================

class SqlAlchemyAttrDictMixin(object):
    """
    Mixin to:

    - get a plain dictionary-like object (with attributes so we can use ``x.y``
      rather than ``x['y']``) from an SQLAlchemy ORM object
    - make a nice ``repr()`` default, maintaining field order

    See https://stackoverflow.com/questions/2537471 and in particular
    https://stackoverflow.com/questions/2441796.
    """

    def get_attrdict(self) -> OrderedNamespace:
        """
        Returns what looks like a plain object with the values of the
        SQLAlchemy ORM object.
        """
        # noinspection PyUnresolvedReferences
        columns = self.__table__.columns.keys()
        values = (getattr(self, x) for x in columns)
        zipped = zip(columns, values)
        return OrderedNamespace(zipped)

    def __repr__(self) -> str:
        return "<{classname}({kvp})>".format(
            classname=type(self).__name__,
            kvp=", ".join(f"{k}={v!r}"
                          for k, v in self.get_attrdict().items())
        )

    @classmethod
    def from_attrdict(cls, attrdict: OrderedNamespace) -> object:
        """
        Builds a new instance of the ORM object from values in an attrdict.
        """
        dictionary = attrdict.__dict__
        # noinspection PyArgumentList
        return cls(**dictionary)


# =============================================================================
# Traverse ORM relationships (SQLAlchemy ORM)
# =============================================================================

def walk_orm_tree(obj,
                  debug: bool = False,
                  seen: Set = None,
                  skip_relationships_always: List[str] = None,
                  skip_relationships_by_tablename: Dict[str, List[str]] = None,
                  skip_all_relationships_for_tablenames: List[str] = None,
                  skip_all_objects_for_tablenames: List[str] = None) \
        -> Generator[object, None, None]:
    """
    Starting with a SQLAlchemy ORM object, this function walks a
    relationship tree, yielding each of the objects once.

    To skip attributes by name, put the attribute name(s) in
    ``skip_attrs_always``. To skip by table name, pass
    ``skip_attrs_by_tablename`` as e.g.

    .. code-block:: python

        {'sometable': ['attr1_to_skip', 'attr2_to_skip']}


    Args:
        obj: the SQLAlchemy ORM object to walk

        debug: be verbose

        seen: usually ``None``, but can be a set of objects marked as "already
            seen"; if an object is in this set, it is skipped

        skip_relationships_always: relationships are skipped if the
            relationship has a name in this (optional) list

        skip_relationships_by_tablename: optional dictionary mapping table
            names (keys) to relationship attribute names (values); if the
            "related table"/"relationship attribute" pair are in this
            dictionary, the relationship is skipped

        skip_all_relationships_for_tablenames: relationships are skipped if the
            the related table has a name in this (optional) list

        skip_all_objects_for_tablenames: if the object belongs to a table whose
            name is in this (optional) list, the object is skipped

    Yields:
        SQLAlchemy ORM objects (including the starting object)

    """
    # http://docs.sqlalchemy.org/en/latest/faq/sessions.html#faq-walk-objects
    skip_relationships_always = skip_relationships_always or []  # type: List[str]  # noqa
    skip_relationships_by_tablename = skip_relationships_by_tablename or {}  # type: Dict[str, List[str]]  # noqa
    skip_all_relationships_for_tablenames = skip_all_relationships_for_tablenames or []  # type: List[str]  # noqa
    skip_all_objects_for_tablenames = skip_all_objects_for_tablenames or []  # type: List[str]  # noqa
    stack = [obj]
    if seen is None:
        seen = set()
    while stack:
        obj = stack.pop(0)
        if obj in seen:
            continue
        tablename = obj.__tablename__
        if tablename in skip_all_objects_for_tablenames:
            continue
        seen.add(obj)
        if debug:
            log.debug("walk: yielding {!r}", obj)
        yield obj
        insp = inspect(obj)  # type: InstanceState
        for relationship in insp.mapper.relationships:  # type: RelationshipProperty  # noqa
            attrname = relationship.key
            # Skip?
            if attrname in skip_relationships_always:
                continue
            if tablename in skip_all_relationships_for_tablenames:
                continue
            if (tablename in skip_relationships_by_tablename and
                    attrname in skip_relationships_by_tablename[tablename]):
                continue
            # Process relationship
            if debug:
                log.debug("walk: following relationship {}", relationship)
            related = getattr(obj, attrname)
            if debug and related:
                log.debug("walk: queueing {!r}", related)
            if relationship.uselist:
                stack.extend(related)
            elif related is not None:
                stack.append(related)


# =============================================================================
# deepcopy an SQLAlchemy object
# =============================================================================
# Use case: object X is in the database; we want to clone it to object Y,
# which we can then save to the database, i.e. copying all SQLAlchemy field
# attributes of X except its PK. We also want it to copy anything that is
# dependent upon X, i.e. traverse relationships.
#
# https://groups.google.com/forum/#!topic/sqlalchemy/wb2M_oYkQdY
# https://groups.google.com/forum/#!searchin/sqlalchemy/cascade%7Csort:date/sqlalchemy/eIOkkXwJ-Ms/JLnpI2wJAAAJ  # noqa

def copy_sqla_object(obj: object,
                     omit_fk: bool = True,
                     omit_pk: bool = True,
                     omit_attrs: List[str] = None,
                     debug: bool = False) -> object:
    """
    Given an SQLAlchemy object, creates a new object (FOR WHICH THE OBJECT
    MUST SUPPORT CREATION USING ``__init__()`` WITH NO PARAMETERS), and copies
    across all attributes, omitting PKs (by default), FKs (by default), and
    relationship attributes (always omitted).

    Args:
        obj: the object to copy
        omit_fk: omit foreign keys (FKs)?
        omit_pk: omit primary keys (PKs)?
        omit_attrs: attributes (by name) not to copy
        debug: be verbose
        
    Returns:
        a new copy of the object
    """
    omit_attrs = omit_attrs or []  # type: List[str]
    cls = type(obj)
    mapper = class_mapper(cls)
    newobj = cls()  # not: cls.__new__(cls)
    rel_keys = set([c.key for c in mapper.relationships])
    prohibited = rel_keys
    if omit_pk:
        pk_keys = set([c.key for c in mapper.primary_key])
        prohibited |= pk_keys
    if omit_fk:
        fk_keys = set([c.key for c in mapper.columns if c.foreign_keys])
        prohibited |= fk_keys
    prohibited |= set(omit_attrs)
    if debug:
        log.debug("copy_sqla_object: skipping: {}", prohibited)
    for k in [p.key for p in mapper.iterate_properties
              if p.key not in prohibited]:
        try:
            value = getattr(obj, k)
            if debug:
                log.debug("copy_sqla_object: processing attribute {} = {}",
                          k, value)
            setattr(newobj, k, value)
        except AttributeError:
            if debug:
                log.debug("copy_sqla_object: failed attribute {}", k)
            pass
    return newobj


def rewrite_relationships(oldobj: object,
                          newobj: object,
                          objmap: Dict[object, object],
                          debug: bool = False,
                          skip_table_names: List[str] = None) -> None:
    """
    A utility function only.
    Used in copying objects between SQLAlchemy sessions.

    Both ``oldobj`` and ``newobj`` are SQLAlchemy instances. The instance
    ``newobj`` is already a copy of ``oldobj`` but we wish to rewrite its
    relationships, according to the map ``objmap``, which maps old to new
    objects.

    For example:

    - Suppose a source session has a Customer record and a Sale record
      containing ``sale.customer_id``, a foreign key to Customer.

    - We may have corresponding Python SQLAlchemy ORM objects
      ``customer_1_src`` and ``sale_1_src``.

    - We copy them into a destination database, where their Python ORM objects
      are ``customer_1_dest`` and ``sale_1_dest``.

    - In the process we set up an object map looking like:

      .. code-block:: none

        Old session     New session
        -------------------------------
        customer_1_src  customer_1_dest
        sale_1_src      sale_1_dest

    - Now, we wish to make ``sale_1_dest`` have a relationship to
      ``customer_1_dest``, in the same way that ``sale_1_src`` has a
      relationship to ``customer_1_src``. This function will modify
      ``sale_1_dest`` accordingly, given this object map. It will observe that
      ``sale_1_src`` (here ``oldobj``) has a relationship to
      ``customer_1_src``; it will note that ``objmap`` maps ``customer_1_src``
      to ``customer_1_dest``; it will create the relationship from
      ``sale_1_dest`` (here ``newobj``) to ``customer_1_dest``.

    Args:
        oldobj: SQLAlchemy ORM object to read from

        newobj: SQLAlchemy ORM object to write to

        objmap: dictionary mapping "source" objects to their corresponding
            "destination" object.

        debug: be verbose

        skip_table_names: if a related table's name is in this (optional) list,
            that relationship is skipped
    """
    skip_table_names = skip_table_names or []  # type: List[str]
    insp = inspect(oldobj)  # type: InstanceState
    # insp.mapper.relationships is of type
    # sqlalchemy.utils._collections.ImmutableProperties, which is basically
    # a sort of AttrDict.
    for attrname_rel in insp.mapper.relationships.items():  # type: Tuple[str, RelationshipProperty]  # noqa
        attrname = attrname_rel[0]
        rel_prop = attrname_rel[1]
        if rel_prop.viewonly:
            if debug:
                log.debug("Skipping viewonly relationship")
            continue  # don't attempt to write viewonly relationships  # noqa
        related_class = rel_prop.mapper.class_
        related_table_name = related_class.__tablename__  # type: str
        if related_table_name in skip_table_names:
            if debug:
                log.debug("Skipping relationship for related table {!r}",
                          related_table_name)
            continue
        # The relationship is an abstract object (so getting the
        # relationship from the old object and from the new, with e.g.
        # newrel = newinsp.mapper.relationships[oldrel.key],
        # yield the same object. All we need from it is the key name.
        #       rel_key = rel.key  # type: str
        # ... but also available from the mapper as attrname, above
        related_old = getattr(oldobj, attrname)
        if rel_prop.uselist:
            related_new = [objmap[r] for r in related_old]
        elif related_old is not None:
            related_new = objmap[related_old]
        else:
            related_new = None
        if debug:
            log.debug("rewrite_relationships: relationship {} -> {}",
                      attrname, related_new)
        setattr(newobj, attrname, related_new)


def deepcopy_sqla_objects(
        startobjs: List[object],
        session: Session,
        flush: bool = True,
        debug: bool = False,
        debug_walk: bool = True,
        debug_rewrite_rel: bool = False,
        objmap: Dict[object, object] = None) -> None:
    """
    Makes a copy of the specified SQLAlchemy ORM objects, inserting them into a
    new session.

    This function operates in several passes:

    1. Walk the ORM tree through all objects and their relationships, copying
       every object thus found (via :func:`copy_sqla_object`, without their
       relationships), and building a map from each source-session object to
       its equivalent destination-session object.

    2. Work through all the destination objects, rewriting their relationships
       (via :func:`rewrite_relationships`) so they relate to each other (rather
       than their source-session brethren).

    3. Insert all the destination-session objects into the destination session.

    For this to succeed, every object must take an ``__init__`` call with no
    arguments (see :func:`copy_sqla_object`). (We can't specify the required
    ``args``/``kwargs``, since we are copying a tree of arbitrary objects.)

    Args:
        startobjs: SQLAlchemy ORM objects to copy
        session: destination SQLAlchemy :class:`Session` into which to insert
            the copies
        flush: flush the session when we've finished?
        debug: be verbose?
        debug_walk: be extra verbose when walking the ORM tree?
        debug_rewrite_rel: be extra verbose when rewriting relationships?
        objmap: starting object map from source-session to destination-session
            objects (see :func:`rewrite_relationships` for more detail);
            usually ``None`` to begin with.
    """
    if objmap is None:
        objmap = {}  # keys = old objects, values = new objects
    if debug:
        log.debug("deepcopy_sqla_objects: pass 1: create new objects")

    # Pass 1: iterate through all objects. (Can't guarantee to get
    # relationships correct until we've done this, since we don't know whether
    # or where the "root" of the PK tree is.)
    seen = set()
    for startobj in startobjs:
        for oldobj in walk_orm_tree(startobj, seen=seen, debug=debug_walk):
            if debug:
                log.debug("deepcopy_sqla_objects: copying {}", oldobj)
            newobj = copy_sqla_object(oldobj, omit_pk=True, omit_fk=True)
            # Don't insert the new object into the session here; it may trigger
            # an autoflush as the relationships are queried, and the new
            # objects are not ready for insertion yet (as their relationships
            # aren't set).
            # Note also the session.no_autoflush option:
            # "sqlalchemy.exc.OperationalError: (raised as a result of Query-
            # invoked autoflush; consider using a session.no_autoflush block if
            # this flush is occurring prematurely)..."
            objmap[oldobj] = newobj

    # Pass 2: set all relationship properties.
    if debug:
        log.debug("deepcopy_sqla_objects: pass 2: set relationships")
    for oldobj, newobj in objmap.items():
        if debug:
            log.debug("deepcopy_sqla_objects: newobj: {}", newobj)
        rewrite_relationships(oldobj, newobj, objmap, debug=debug_rewrite_rel)

    # Now we can do session insert.
    if debug:
        log.debug("deepcopy_sqla_objects: pass 3: insert into session")
    for newobj in objmap.values():
        session.add(newobj)

    # Done
    if debug:
        log.debug("deepcopy_sqla_objects: done")
    if flush:
        session.flush()


def deepcopy_sqla_object(startobj: object,
                         session: Session,
                         flush: bool = True,
                         debug: bool = False,
                         debug_walk: bool = False,
                         debug_rewrite_rel: bool = False,
                         objmap: Dict[object, object] = None) -> object:
    """
    Makes a copy of the object, inserting it into ``session``.

    Uses :func:`deepcopy_sqla_objects` (q.v.).

    A problem is the creation of duplicate dependency objects if you call it
    repeatedly.

    Optionally, if you pass the objmap in (which maps old to new objects), you
    can call this function repeatedly to clone a related set of objects...
    ... no, that doesn't really work, as it doesn't visit parents before
    children. The :func:`cardinal_pythonlib.sqlalchemy.merge_db.merge_db`
    function does that properly.

    Args:
        startobj: SQLAlchemy ORM object to deep-copy
        session: see :func:`deepcopy_sqla_objects`
        flush: see :func:`deepcopy_sqla_objects`
        debug: see :func:`deepcopy_sqla_objects`
        debug_walk: see :func:`deepcopy_sqla_objects`
        debug_rewrite_rel: see :func:`deepcopy_sqla_objects`
        objmap: see :func:`deepcopy_sqla_objects`

    Returns:
        the copied object matching ``startobj``

    """
    if objmap is None:
        objmap = {}  # keys = old objects, values = new objects
    deepcopy_sqla_objects(
        startobjs=[startobj],
        session=session,
        flush=flush,
        debug=debug,
        debug_walk=debug_walk,
        debug_rewrite_rel=debug_rewrite_rel,
        objmap=objmap
    )
    return objmap[startobj]  # returns the new object matching startobj


# =============================================================================
# Get Columns from an ORM instance
# =============================================================================

def gen_columns(obj) -> Generator[Tuple[str, Column], None, None]:
    """
    Asks a SQLAlchemy ORM object: "what are your SQLAlchemy columns?"

    Yields tuples of ``(attr_name, Column)`` from an SQLAlchemy ORM object
    instance. Also works with the corresponding SQLAlchemy ORM class. Examples:

    .. code-block:: python

        from sqlalchemy.ext.declarative import declarative_base
        from sqlalchemy.sql.schema import Column
        from sqlalchemy.sql.sqltypes import Integer

        Base = declarative_base()

        class MyClass(Base):
            __tablename__ = "mytable"
            pk = Column("pk", Integer, primary_key=True, autoincrement=True)
            a = Column("a", Integer)

        x = MyClass()

        list(gen_columns(x))
        list(gen_columns(MyClass))

    """
    mapper = obj.__mapper__  # type: Mapper
    assert mapper, (
        f"gen_columns called on {obj!r} which is not an SQLAlchemy ORM object"
    )
    colmap = mapper.columns  # type: OrderedProperties
    if not colmap:
        return
    for attrname, column in colmap.items():
        # NB: column.name is the SQL column name, not the attribute name
        yield attrname, column
    # Don't bother using
    #   cls = obj.__class_
    #   for attrname in dir(cls):
    #       cls_attr = getattr(cls, attrname)
    #       # ... because, for columns, these will all be instances of
    #       # sqlalchemy.orm.attributes.InstrumentedAttribute.


def get_pk_attrnames(obj) -> List[str]:
    """
    Asks an SQLAlchemy ORM object: "what are your primary key(s)?"

    Args:
        obj: SQLAlchemy ORM object

    Returns:
        list of attribute names of primary-key columns

    """
    return [attrname
            for attrname, column in gen_columns(obj)
            if column.primary_key]


def gen_columns_for_uninstrumented_class(cls: Type) \
        -> Generator[Tuple[str, Column], None, None]:
    """
    Generate ``(attr_name, Column)`` tuples from an UNINSTRUMENTED class, i.e.
    one that does not inherit from ``declarative_base()``. Use this for mixins
    of that kind.

    SUBOPTIMAL. May produce warnings like:
    
    .. code-block:: none
    
        SAWarning: Unmanaged access of declarative attribute id from non-mapped class GenericTabletRecordMixin

    Try to use :func:`gen_columns` instead.
    """  # noqa
    for attrname in dir(cls):
        potential_column = getattr(cls, attrname)
        if isinstance(potential_column, Column):
            yield attrname, potential_column


def attrname_to_colname_dict(cls) -> Dict[str, str]:
    """
    Asks an SQLAlchemy class how its attribute names correspond to database
    column names.

    Args:
        cls: SQLAlchemy ORM class

    Returns:
        a dictionary mapping attribute names to database column names
    """
    attr_col = {}  # type: Dict[str, str]
    for attrname, column in gen_columns(cls):
        attr_col[attrname] = column.name
    return attr_col


def colname_to_attrname_dict(cls) -> Dict[str, str]:
    return reversedict(attrname_to_colname_dict(cls))


# =============================================================================
# Get relationships from an ORM instance
# =============================================================================

def gen_relationships(obj) -> Generator[Tuple[str, RelationshipProperty, Type],
                                        None, None]:
    """
    Yields tuples of ``(attrname, RelationshipProperty, related_class)``
    for all relationships of an ORM object.
    The object 'obj' can be EITHER an instance OR a class.
    """
    insp = inspect(obj)  # type: InstanceState
    # insp.mapper.relationships is of type
    # sqlalchemy.utils._collections.ImmutableProperties, which is basically
    # a sort of AttrDict.
    for attrname, rel_prop in insp.mapper.relationships.items():  # type: Tuple[str, RelationshipProperty]  # noqa
        # noinspection PyUnresolvedReferences
        related_class = rel_prop.mapper.class_
        # log.critical("gen_relationships: attrname={!r}, "
        #              "rel_prop={!r}, related_class={!r}, rel_prop.info={!r}",
        #              attrname, rel_prop, related_class, rel_prop.info)
        yield attrname, rel_prop, related_class


# =============================================================================
# Inspect ORM objects (SQLAlchemy ORM)
# =============================================================================

def get_orm_columns(cls: Type) -> List[Column]:
    """
    Gets :class:`Column` objects from an SQLAlchemy ORM class.
    Does not provide their attribute names.
    """
    mapper = inspect(cls)  # type: Mapper
    # ... returns InstanceState if called with an ORM object
    #     http://docs.sqlalchemy.org/en/latest/orm/session_state_management.html#session-object-states  # noqa
    # ... returns Mapper if called with an ORM class
    #     http://docs.sqlalchemy.org/en/latest/orm/mapping_api.html#sqlalchemy.orm.mapper.Mapper  # noqa
    colmap = mapper.columns  # type: OrderedProperties
    return colmap.values()


def get_orm_column_names(cls: Type, sort: bool = False) -> List[str]:
    """
    Gets column names (that is, database column names) from an SQLAlchemy
    ORM class.
    """
    colnames = [col.name for col in get_orm_columns(cls)]
    return sorted(colnames) if sort else colnames


# =============================================================================
# Inspect metadata (SQLAlchemy ORM)
# =============================================================================

def get_table_names_from_metadata(metadata: MetaData) -> List[str]:
    """
    Returns all database table names found in an SQLAlchemy :class:`MetaData`
    object.
    """
    return [table.name for table in metadata.tables.values()]


def get_metadata_from_orm_class_or_object(cls: Type) -> MetaData:
    """
    Returns the :class:`MetaData` object from an SQLAlchemy ORM class or
    instance.
    """
    # noinspection PyUnresolvedReferences
    table = cls.__table__  # type: Table
    return table.metadata


def gen_orm_classes_from_base(base: Type) -> Generator[Type, None, None]:
    """
    From an SQLAlchemy ORM base class, yield all the subclasses (except those
    that are abstract).

    If you begin with the proper :class`Base` class, then this should give all
    ORM classes in use.
    """
    for cls in gen_all_subclasses(base):
        if _get_immediate_cls_attr(cls, '__abstract__', strict=True):
            # This is SQLAlchemy's own way of detecting abstract classes; see
            # sqlalchemy.ext.declarative.base
            continue  # NOT an ORM class
        yield cls


def get_orm_classes_by_table_name_from_base(base: Type) -> Dict[str, Type]:
    """
    Given an SQLAlchemy ORM base class, returns a dictionary whose keys are
    table names and whose values are ORM classes.

    If you begin with the proper :class`Base` class, then this should give all
    tables and ORM classes in use.
    """
    # noinspection PyUnresolvedReferences
    return {cls.__tablename__: cls for cls in gen_orm_classes_from_base(base)}



#!/usr/bin/env python
# cardinal_pythonlib/reprfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to assist making repr() methods for Python objects.**

"""

import pprint
from typing import Any, Iterable, List, Tuple

COMMA_SPACE = ", "


# =============================================================================
# __repr__ aids
# =============================================================================
# The repr() function often attempts to return something suitable for eval();
# failing that, it usually shows an address.
# https://docs.python.org/3/library/functions.html#repr

def repr_result(obj: Any, elements: List[str],
                with_addr: bool = False, joiner: str = COMMA_SPACE) -> str:
    """
    Internal function to make a :func:`repr`-style representation of an object.

    Args:
        obj: object to display
        elements: list of object ``attribute=value`` strings
        with_addr: include the memory address of ``obj``
        joiner: string with which to join the elements

    Returns:
        string: :func:`repr`-style representation

    """
    if with_addr:
        return "<{qualname}({elements}) at {addr}>".format(
            qualname=obj.__class__.__qualname__,
            elements=joiner.join(elements),
            addr=hex(id(obj)))
    else:
        return "{qualname}({elements})".format(
            qualname=obj.__class__.__qualname__,
            elements=joiner.join(elements))


def auto_repr(obj: Any, with_addr: bool = False,
              sort_attrs: bool = True, joiner: str = COMMA_SPACE) -> str:
    """
    Convenience function for :func:`__repr__`.
    Works its way through the object's ``__dict__`` and reports accordingly.

    Args:
        obj: object to display
        with_addr: include the memory address of ``obj``
        sort_attrs: sort the attributes into alphabetical order?
        joiner: string with which to join the elements

    Returns:
        string: :func:`repr`-style representation
    """
    if sort_attrs:
        keys = sorted(obj.__dict__.keys())
    else:
        keys = obj.__dict__.keys()
    elements = [f"{k}={getattr(obj, k)!r}" for k in keys]
    return repr_result(obj, elements, with_addr=with_addr, joiner=joiner)


def simple_repr(obj: Any, attrnames: List[str],
                with_addr: bool = False, joiner: str = COMMA_SPACE) -> str:
    """
    Convenience function for :func:`__repr__`.
    Works its way through a list of attribute names, and creates a ``repr()``
    representation assuming that parameters to the constructor have the same
    names.

    Args:
        obj: object to display
        attrnames: names of attributes to include
        with_addr: include the memory address of ``obj``
        joiner: string with which to join the elements

    Returns:
        string: :func:`repr`-style representation

    """
    elements = [f"{name}={getattr(obj, name)!r}" for name in attrnames]
    return repr_result(obj, elements, with_addr=with_addr, joiner=joiner)


def mapped_repr(obj: Any, attributes: List[Tuple[str, str]],
                with_addr: bool = False, joiner: str = COMMA_SPACE) -> str:
    """
    Convenience function for :func:`__repr__`.
    Takes attribute names and corresponding initialization parameter names
    (parameters to :func:`__init__`).

    Args:
        obj: object to display
        attributes: list of tuples, each ``(attr_name, init_param_name)``.
        with_addr: include the memory address of ``obj``
        joiner: string with which to join the elements

    Returns:
        string: :func:`repr`-style representation

    """
    elements = [f"{init_param_name}={getattr(obj, attr_name)!r}"
                for attr_name, init_param_name in attributes]
    return repr_result(obj, elements, with_addr=with_addr, joiner=joiner)


def mapped_repr_stripping_underscores(
        obj: Any, attrnames: List[str],
        with_addr: bool = False, joiner: str = COMMA_SPACE) -> str:
    """
    Convenience function for :func:`__repr__`.
    Here, you pass a list of internal attributes, and it assumes that the
    :func:`__init__` parameter names have the leading underscore dropped.

    Args:
        obj: object to display
        attrnames: list of attribute names
        with_addr: include the memory address of ``obj``
        joiner: string with which to join the elements

    Returns:
        string: :func:`repr`-style representation

    """
    attributes = []
    for attr_name in attrnames:
        if attr_name.startswith('_'):
            init_param_name = attr_name[1:]
        else:
            init_param_name = attr_name
        attributes.append((attr_name, init_param_name))
    return mapped_repr(obj, attributes, with_addr=with_addr, joiner=joiner)


def ordered_repr(obj: object, attrlist: Iterable[str],
                 joiner: str = COMMA_SPACE) -> str:
    """
    Shortcut to make :func:`repr` functions ordered.
    Define your :func:`__repr__` like this:

    .. code-block:: python

        def __repr__(self):
            return ordered_repr(self, ["field1", "field2", "field3"])

    Args:
        obj: object to display
        attrlist: iterable of attribute names
        joiner: string with which to join the elements

    Returns:
        string: :func:`repr`-style representation
    """
    return "<{classname}({kvp})>".format(
        classname=type(obj).__name__,
        kvp=joiner.join(f"{a}={getattr(obj, a)!r}" for a in attrlist)
    )


def auto_str(obj: Any, indent: int = 4, width: int = 80, depth: int = None,
             compact: bool = False) -> str:
    """
    Make a pretty :func:`str()` representation using :func:`pprint.pformat`
    and the object's ``__dict__`` attribute.

    Args:
        obj: object to display
        indent: see
            https://docs.python.org/3/library/pprint.html#pprint.PrettyPrinter
        width: as above
        depth: as above
        compact: as above

    Returns:
        string: :func:`str`-style representation

    """
    return pprint.pformat(obj.__dict__, indent=indent, width=width,
                          depth=depth, compact=compact)



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/subproc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================
"""

import re
from typing import Any, Dict, Iterable, List
import unicodedata


# =============================================================================
# Finding
# =============================================================================

def find_nth(s: str, x: str, n: int = 0, overlap: bool = False) -> int:
    """
    Finds the position of *n*\ th occurrence of ``x`` in ``s``, or ``-1`` if
    there isn't one.

    - The ``n`` parameter is zero-based (i.e. 0 for the first, 1 for the
      second...).
    - If ``overlap`` is true, allows fragments to overlap. If not, they must be
      distinct.

    As per
    https://stackoverflow.com/questions/1883980/find-the-nth-occurrence-of-substring-in-a-string
    """  # noqa
    length_of_fragment = 1 if overlap else len(x)
    i = -length_of_fragment
    for _ in range(n + 1):
        i = s.find(x, i + length_of_fragment)
        if i < 0:
            break
    return i


# =============================================================================
# Splitting
# =============================================================================

def split_string(x: str, n: int) -> List[str]:
    """
    Split string into chunks of length n
    """
    # https://stackoverflow.com/questions/9475241/split-string-every-nth-character  # noqa
    return [x[i:i+n] for i in range(0, len(x), n)]


# =============================================================================
# Replacement
# =============================================================================

def multiple_replace(text: str, rep: Dict[str, str]) -> str:
    """
    Returns a version of ``text`` in which the keys of ``rep`` (a dict) have
    been replaced by their values.

    As per
    https://stackoverflow.com/questions/6116978/python-replace-multiple-strings.
    """
    rep = dict((re.escape(k), v) for k, v in rep.items())
    pattern = re.compile("|".join(rep.keys()))
    return pattern.sub(lambda m: rep[re.escape(m.group(0))], text)


def replace_in_list(stringlist: Iterable[str],
                    replacedict: Dict[str, str]) -> List[str]:
    """
    Returns a list produced by applying :func:`multiple_replace` to every
    string in ``stringlist``.

    Args:
        stringlist: list of source strings
        replacedict: dictionary mapping "original" to "replacement" strings

    Returns:
        list of final strings

    """
    newlist = []
    for fromstring in stringlist:
        newlist.append(multiple_replace(fromstring, replacedict))
    return newlist


# =============================================================================
# Mangling to ASCII
# =============================================================================

def mangle_unicode_to_ascii(s: Any) -> str:
    """
    Mangle unicode to ASCII, losing accents etc. in the process.
    """
    # https://stackoverflow.com/questions/1207457
    if s is None:
        return ""
    if not isinstance(s, str):
        s = str(s)
    return (
        unicodedata.normalize('NFKD', s)
                   .encode('ascii', 'ignore')  # gets rid of accents
                   .decode('ascii')  # back to a string
    )


# =============================================================================
# Making strings and string lists
# =============================================================================

def strnum(prefix: str, num: int, suffix: str = "") -> str:
    """
    Makes a string of the format ``<prefix><number><suffix>``.
    """
    return f"{prefix}{num}{suffix}"


def strnumlist(prefix: str, numbers: List[int], suffix: str = "") -> List[str]:
    """
    Makes a string of the format ``<prefix><number><suffix>`` for every number
    in ``numbers``, and returns them as a list.
    """
    return [f"{prefix}{num}{suffix}" for num in numbers]


def strseq(prefix: str, first: int, last: int, suffix: str = "") -> List[str]:
    """
    Makes a string of the format ``<prefix><number><suffix>`` for every number
    from ``first`` to ``last`` inclusive, and returns them as a list.
    """
    return [strnum(prefix, n, suffix) for n in range(first, last + 1)]



#!/usr/bin/env python
# cardinal_pythonlib/progress.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Counters etc. for visual display of progress.**

"""

import logging

log = logging.getLogger(__name__)


# =============================================================================
# Visuals
# =============================================================================

class ActivityCounter(object):
    """
    Simple class to report progress in a repetitive activity.
    """
    def __init__(self,
                 activity: str,
                 n_total: int = None,
                 report_every: int = 1000,
                 loglevel: int = logging.DEBUG) -> None:
        """
        Args:
            activity:
                Description of the repetitive activity being performed.
            n_total:
                If known, the total number of iterations required.
            report_every:
                Report progress every n operations.
            loglevel:
                Log level to use.
        """
        self.activity = activity
        self.count = 0
        self.n_total = n_total
        self.report_every = report_every
        self.loglevel = loglevel

    def tick(self) -> None:
        """
        Note a further occurrence, and report progress if required.
        """
        self.count += 1
        c = self.count
        n = self.n_total
        if c == 1 or c % self.report_every == 0 or c == n:
            if self.n_total is not None:
                of_n = f" of {n}"
            else:
                of_n = ""
            log.log(self.loglevel, f"{self.activity} {c}{of_n}")



#!/usr/bin/env python
# cardinal_pythonlib/crypto.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions involving cryptography.**

"""

# The following requires a C compiler, so we don't have it in our standard
# requirements. However, it is vital for this module.
#
# noinspection PyUnresolvedReferences
import bcrypt  # pip install bcrypt; see https://pypi.org/project/bcrypt/


# =============================================================================
# bcrypt
# =============================================================================

BCRYPT_DEFAULT_LOG_ROUNDS = 12  # bcrypt default; work factor is 2^this.


def hash_password(plaintextpw: str,
                  log_rounds: int = BCRYPT_DEFAULT_LOG_ROUNDS) -> str:
    """
    Makes a hashed password (using a new salt) using ``bcrypt``.

    The hashed password includes the salt at its start, so no need to store a
    separate salt.
    """
    salt = bcrypt.gensalt(log_rounds)  # optional parameter governs complexity
    hashedpw = bcrypt.hashpw(plaintextpw, salt)
    return hashedpw


def is_password_valid(plaintextpw: str, storedhash: str) -> bool:
    """
    Checks if a plaintext password matches a stored hash.

    Uses ``bcrypt``. The stored hash includes its own incorporated salt.
    """
    # Upon CamCOPS from MySQL 5.5.34 (Ubuntu) to 5.1.71 (CentOS 6.5), the
    # VARCHAR was retrieved as Unicode. We needed to convert that to a str.
    # For Python 3 compatibility, we just str-convert everything, avoiding the
    # unicode keyword, which no longer exists.
    if storedhash is None:
        storedhash = ""
    storedhash = str(storedhash)
    if plaintextpw is None:
        plaintextpw = ""
    plaintextpw = str(plaintextpw)
    try:
        h = bcrypt.hashpw(plaintextpw, storedhash)
    except ValueError:  # e.g. ValueError: invalid salt
        return False
    return h == storedhash



#!/usr/bin/env python
# cardinal_pythonlib/betweendict.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Implements BetweenDict.***

"""

from typing import Dict


# =============================================================================
# Range dictionary for comparisons
# =============================================================================

class BetweenDict(dict):
    # noinspection HttpUrlsUsage
    """
    Range dictionary for comparisons.

    - Create it with **keys** that are tuples/lists of length 2, such as
      ``(lower, upper)``, and **values** that are the values
      that this particular range should be mapped to.
    - Access it using a specific test value as the key, and the result will be
      the value for an appropriate range (or :exc:`KeyError` will be raised if
      none of the ranges contain the test value).
    - The test is ``lower <= testvalue < upper``, i.e. ranges of ``[lower,
      upper)``.

    Example:

    .. code-block:: python

        bd = BetweenDict({
            # tests a <= x < b
            (100, float("inf")): "very high",  # from Python 3.5, can use math.inf
            (75, 100): "high",
            (50, 75): "medium",
            (25, 50): "low",
            (0, 25): "very low",
        })

        bd[500]  # 'very high'
        bd[80]  # 'high'
        bd[50]  # 'medium'
        bd[-5]  # raises KeyError

    Various implementation alternatives:

    - http://joshuakugler.com/archives/30-BetweenDict,-a-Python-dict-for-value-ranges.html
      ... NB has initialization default argument bug
    - https://pypi.python.org/pypi/rangedict/0.1.5
    - https://stackoverflow.com/questions/30254739/is-there-a-library-implemented-rangedict-in-python
    """  # noqa
    INVALID_MSG_TYPE = "Key must be an iterable with length 2"
    INVALID_MSG_VALUE = "First element of key must be less than second element"

    # noinspection PyMissingConstructor
    def __init__(self, d: Dict = None) -> None:
        d = d or {}
        for k, v in d.items():
            self[k] = v

    def __getitem__(self, key):
        for k, v in self.items():
            if k[0] <= key < k[1]:
                return v
        raise KeyError(f"Key {key!r} is not in any ranges")

    def __setitem__(self, key, value):
        try:
            if len(key) != 2:
                raise ValueError(self.INVALID_MSG_TYPE)
        except TypeError:
            raise TypeError(self.INVALID_MSG_TYPE)
        if key[0] < key[1]:
            super().__setitem__((key[0], key[1]), value)
        else:
            raise RuntimeError(self.INVALID_MSG_VALUE)

    def __contains__(self, key):
        try:
            # noinspection PyStatementEffect
            self[key]
            return True
        except KeyError:
            return False



#!/usr/bin/env python
# cardinal_pythonlib/athena_ohdsi.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to assist with SNOMED-CT.**

See https://www.snomed.org/.

Note that the licensing arrangements for SNOMED-CT mean that the actual codes
must be separate (and not part of this code).

A full SNOMED CT download is about 1.1 Gb; see
https://digital.nhs.uk/services/terminology-and-classifications/snomed-ct.
Within a file such as ``uk_sct2cl_26.0.2_20181107000001.zip``, relevant files
include:

.. code-block:: none

    # Files with "Amoxicillin" in include two snapshots and two full files:

    SnomedCT_UKClinicalRF2_PRODUCTION_20181031T000001Z/Full/Terminology/sct2_Description_Full-en-GB_GB1000000_20181031.txt
    # ... 234,755 lines

    SnomedCT_InternationalRF2_PRODUCTION_20180731T120000Z/Full/Terminology/sct2_Description_Full-en_INT_20180731.txt
    # ... 2,513,953 lines; this is the main file.

Note grammar:

- https://www.snomed.org/scg
- https://confluence.ihtsdotools.org/display/DOCSCG
- https://confluence.ihtsdotools.org/download/attachments/33494865/SnomedCtExpo_Expressions_20161028_s2_20161101.pdf  # noqa
- https://confluence.ihtsdotools.org/display/SLPG/SNOMED+CT+Expression+Constraint+Language

Test basic expressions:

.. code-block:: python

    import logging
    from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger
    from cardinal_pythonlib.snomed import *
    main_only_quicksetup_rootlogger(level=logging.DEBUG)

    # ---------------------------------------------------------------------
    # From the SNOMED-CT examples (https://www.snomed.org/scg), with some
    # values fixed from the term browser:
    # ---------------------------------------------------------------------

    diabetes = SnomedConcept(73211009, "Diabetes mellitus (disorder)")
    diabetes_expr = SnomedExpression(diabetes)
    print(diabetes_expr.longform)
    print(diabetes_expr.shortform)

    pain = SnomedConcept(22253000, "Pain (finding)")
    finding_site = SnomedConcept(36369800, "Finding site")
    foot = SnomedConcept(56459004, "Foot")

    pain_in_foot = SnomedExpression(pain, {finding_site: foot})
    print(pain_in_foot.longform)
    print(pain_in_foot.shortform)

    amoxicillin_medicine = SnomedConcept(27658006, "Product containing amoxicillin (medicinal product)")
    amoxicillin_substance = SnomedConcept(372687004, "Amoxicillin (substance)")
    has_dose_form = SnomedConcept(411116001, "Has manufactured dose form (attribute)")
    capsule = SnomedConcept(385049006, "Capsule (basic dose form)")
    has_active_ingredient = SnomedConcept(127489000, "Has active ingredient (attribute)")
    has_basis_of_strength_substance = SnomedConcept(732943007, "Has basis of strength substance (attribute)")
    mass = SnomedConcept(118538004, "Mass, a measure of quantity of matter (property) (qualifier value)")
    unit_of_measure = SnomedConcept(767524001, "Unit of measure (qualifier value)")
    milligrams = SnomedConcept(258684004, "milligram (qualifier value)")

    amoxicillin_500mg_capsule = SnomedExpression(
        amoxicillin_medicine, [
            SnomedAttributeSet({has_dose_form: capsule}),
            SnomedAttributeGroup({
                has_active_ingredient: amoxicillin_substance,
                has_basis_of_strength_substance: SnomedExpression(
                    amoxicillin_substance, {
                        mass: 500,
                        unit_of_measure: milligrams,
                    }
                ),
            }),
        ]
    )
    print(amoxicillin_500mg_capsule.longform)
    print(amoxicillin_500mg_capsule.shortform)

"""  # noqa

from typing import Dict, Iterable, List, Union

from cardinal_pythonlib.reprfunc import simple_repr


# =============================================================================
# Constants
# =============================================================================

BACKSLASH = "\\"
COLON = ":"
COMMA = ","
EQUALS = "="
HASH = "#"
LBRACE = "{"
LBRACKET = "("
PIPE = "|"
PLUS = "+"
QM = '"'  # double quotation mark
RBRACE = "}"
RBRACKET = ")"
TAB = "\t"
NEWLINE = "\n"

ID_MIN_DIGITS = 6
ID_MAX_DIGITS = 18

VALUE_TYPE = Union["SnomedConcept", "SnomedExpression", int, float, str]
DICT_ATTR_TYPE = Dict["SnomedConcept", VALUE_TYPE]

SNOMED_XML_NAME = "snomed_ct_expression"


# =============================================================================
# Quoting strings
# =============================================================================

def double_quoted(s: str) -> str:
    r"""
    Returns a representation of the string argument with double quotes and
    escaped characters.

    Args:
        s: the argument

    See:

    - https://code.activestate.com/lists/python-list/272714/ -- does not work
      as null values get escaped in different ways in modern Python, and in a
      slightly unpredictable way
    - https://mail.python.org/pipermail/python-list/2003-April/236940.html --
      won't deal with repr() using triple-quotes
    - https://stackoverflow.com/questions/1675181/get-str-repr-with-double-quotes-python
      -- probably the right general approach

    Test code:

    .. code-block:: python

        from cardinal_pythonlib.snomed import double_quoted

        def test(s):
            print(f"double_quoted({s!r}) -> {double_quoted(s)}")


        test("ab'cd")
        test("ab'c\"d")
        test('ab"cd')

    """  # noqa
    # For efficiency, we use a list:
    # https://stackoverflow.com/questions/3055477/how-slow-is-pythons-string-concatenation-vs-str-join  # noqa
    # https://waymoot.org/home/python_string/
    dquote = '"'
    ret = [dquote]  # type: List[str]
    for c in s:
        # "Named" characters
        if c == NEWLINE:
            ret.append(r"\n")
        elif c == TAB:
            ret.append(r"\t")
        elif c == QM:
            ret.append(r'\"')
        elif c == BACKSLASH:
            ret.append(r"\\")
        elif ord(c) < 32:
            # two-digit hex format, e.g. \x1F for ASCII 31
            ret.append(fr"\x{ord(c):02X}")
        else:
            ret.append(c)
    ret.append(dquote)
    return "".join(ret)


# =============================================================================
# SNOMED-CT concepts
# =============================================================================

class SnomedBase(object):
    """
    Common functions for SNOMED-CT classes
    """
    def as_string(self, longform: bool = True) -> str:
        """
        Returns the string form.

        Args:
            longform: print SNOMED-CT concepts in long form?
        """
        raise NotImplementedError("implement in subclass")

    @property
    def shortform(self) -> str:
        """
        Returns the short form, without terms.
        """
        return self.as_string(False)

    @property
    def longform(self) -> str:
        return self.as_string(True)

    def __str__(self) -> str:
        return self.as_string(True)


class SnomedConcept(SnomedBase):
    """
    Represents a SNOMED concept with its description (associated term).
    """
    def __init__(self, identifier: int, term: str) -> None:
        """
        Args:
            identifier: SNOMED-CT identifier (code)
            term: associated term (description)
        """
        assert isinstance(identifier, int), (
            f"SNOMED-CT concept identifier is not an integer: {identifier!r}"
        )
        ndigits = len(str(identifier))
        assert ID_MIN_DIGITS <= ndigits <= ID_MAX_DIGITS, (
            f"SNOMED-CT concept identifier has wrong number of digits: "
            f"{identifier!r}"
        )
        assert PIPE not in term, (
            f"SNOMED-CT term has invalid pipe character: {term!r}"
        )
        self.identifier = identifier
        self.term = term

    def __repr__(self) -> str:
        return simple_repr(self, ["identifier", "term"])

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class.
        if longform:
            return f"{self.identifier} {PIPE}{self.term}{PIPE}"
        else:
            return str(self.identifier)

    def concept_reference(self, longform: bool = True) -> str:
        """
        Returns one of the string representations.

        Args:
            longform: in long form, with the description (associated term)?
        """
        return self.as_string(longform)


# =============================================================================
# SNOMED-CT expressions
# =============================================================================

class SnomedValue(SnomedBase):
    """
    Represents a value: either a concrete value (e.g. int, float, str), or a
    SNOMED-CT concept/expression.

    Implements the grammar elements: attributeValue, expressionValue,
    stringValue, numericValue, integerValue, decimalValue.
    """
    def __init__(self, value: VALUE_TYPE) -> None:
        """
        Args:
            value: the value
        """
        assert isinstance(value, (SnomedConcept, SnomedExpression,
                                  int, float, str)), (
            f"Invalid value type to SnomedValue: {value!r}"
        )
        self.value = value

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class
        x = self.value
        if isinstance(x, SnomedConcept):
            return x.concept_reference(longform)
        elif isinstance(x, SnomedExpression):
            # As per p16 of formal reference cited above.
            return f"{LBRACKET} {x.as_string(longform)} {RBRACKET}"
        elif isinstance(x, (int, float)):
            return HASH + str(x)
        elif isinstance(x, str):
            # On the basis that SNOMED's "QM" (quote mark) is 0x22, the double
            # quote:
            return double_quoted(x)
        else:
            raise ValueError("Bad input value type")

    def __repr__(self) -> str:
        return simple_repr(self, ["value"])


class SnomedFocusConcept(SnomedBase):
    """
    Represents a SNOMED-CT focus concept, which is one or more concepts.
    """
    def __init__(self,
                 concept: Union[SnomedConcept, Iterable[SnomedConcept]]) \
            -> None:
        """
        Args:
            concept: the core concept(s); a :class:`SnomedCode` or an
                iterable of them
        """
        if isinstance(concept, SnomedConcept):
            self.concepts = [concept]
        else:
            self.concepts = list(concept)
        assert all(isinstance(x, SnomedConcept) for x in self.concepts)

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class.
        sep = " " + PLUS + " "
        return sep.join(c.concept_reference(longform) for c in self.concepts)

    def __repr__(self) -> str:
        return simple_repr(self, ["concepts"])


class SnomedAttribute(SnomedBase):
    """
    Represents a SNOMED-CT attribute, being a name/value pair.
    """
    def __init__(self, name: SnomedConcept, value: VALUE_TYPE) -> None:
        """
        Args:
            name: a :class:`SnomedConcept` (attribute name)
            value: an attribute value (:class:`SnomedConcept`, number, or
                string)
        """
        assert isinstance(name, SnomedConcept)
        if not isinstance(value, SnomedValue):
            value = SnomedValue(value)
        self.name = name
        self.value = value

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class.
        return (
            f"{self.name.concept_reference(longform)} {EQUALS} "
            f"{self.value.as_string(longform)}"
        )

    def __repr__(self) -> str:
        return simple_repr(self, ["name", "value"])


class SnomedAttributeSet(SnomedBase):
    """
    Represents an attribute set.
    """
    def __init__(self, attributes: Union[DICT_ATTR_TYPE,
                                         Iterable[SnomedAttribute]]) -> None:
        """
        Args:
            attributes: the attributes
        """
        if isinstance(attributes, dict):
            self.attributes = [SnomedAttribute(k, v)
                               for k, v in attributes.items()]
        else:
            self.attributes = list(attributes)
        assert all(isinstance(x, SnomedAttribute) for x in self.attributes)

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class.
        attrsep = COMMA + " "
        return attrsep.join(attr.as_string(longform)
                            for attr in self.attributes)

    def __repr__(self) -> str:
        return simple_repr(self, ["attributes"])


class SnomedAttributeGroup(SnomedBase):
    """
    Represents a collected group of attribute/value pairs.
    """
    def __init__(self, attribute_set: Union[DICT_ATTR_TYPE,
                                            SnomedAttributeSet]) -> None:
        """
        Args:
            attribute_set: a :class:`SnomedAttributeSet` to group
        """
        if isinstance(attribute_set, dict):
            attribute_set = SnomedAttributeSet(attribute_set)
        assert isinstance(attribute_set, SnomedAttributeSet)
        self.attribute_set = attribute_set

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class.
        return f"{LBRACE} {self.attribute_set.as_string(longform)} {RBRACE}"

    def __repr__(self) -> str:
        return simple_repr(self, ["attribute_set"])


class SnomedRefinement(SnomedBase):
    """
    Implements a SNOMED-CT "refinement", which is an attribute set +/- some
    attribute groups.
    """
    def __init__(self,
                 refinements: Union[DICT_ATTR_TYPE,
                                    Iterable[Union[SnomedAttributeSet,
                                                   SnomedAttributeGroup]]]) \
            -> None:
        """
        Args:
            refinements: iterable of :class:`SnomedAttributeSet` (but only
                zero or one) and :class:`SnomedAttributeGroup` objects
        """
        if isinstance(refinements, dict):
            refinements = [SnomedAttributeSet(refinements)]
        self.attrsets = []  # type: List[SnomedBase]
        self.attrgroups = []  # type: List[SnomedBase]
        for r in refinements:
            if isinstance(r, SnomedAttributeSet):
                if self.attrsets:
                    raise ValueError("Only one SnomedAttributeSet allowed "
                                     "to SnomedRefinement")
                self.attrsets.append(r)
            elif isinstance(r, SnomedAttributeGroup):
                self.attrgroups.append(r)
            else:
                raise ValueError(f"Unknown object to SnomedRefinement: {r!r}")

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class.
        # Ungrouped before grouped; see 6.5 in "SNOMED CT Compositional Grammar
        # v2.3.1"
        sep = COMMA + " "
        return sep.join(x.as_string(longform)
                        for x in self.attrsets + self.attrgroups)

    def __repr__(self) -> str:
        return simple_repr(self, ["attrsets", "attrgroups"])


class SnomedExpression(SnomedBase):
    """
    An expression containing several SNOMED-CT codes in relationships.
    """
    def __init__(self,
                 focus_concept: Union[SnomedConcept, SnomedFocusConcept],
                 refinement: Union[SnomedRefinement,
                                   DICT_ATTR_TYPE,
                                   List[Union[SnomedAttributeSet,
                                              SnomedAttributeGroup]]] = None) \
            -> None:
        """
        Args:
            focus_concept: the core concept(s); a :class:`SnomedFocusConcept`
            refinement: optional additional information; a
                :class:`SnomedRefinement` or a dictionary or list that can be
                converted to one
        """
        if isinstance(focus_concept, SnomedConcept):
            focus_concept = SnomedFocusConcept(focus_concept)
        assert isinstance(focus_concept, SnomedFocusConcept)
        if isinstance(refinement, (dict, list)):
            refinement = SnomedRefinement(refinement)
        if refinement is not None:
            assert isinstance(refinement, SnomedRefinement)
        self.focus_concept = focus_concept
        self.refinement = refinement

    def as_string(self, longform: bool = True) -> str:
        # Docstring in base class.
        s = self.focus_concept.as_string(longform)
        if self.refinement:
            s += " " + COLON + " " + self.refinement.as_string(longform)
        return s

    def __repr__(self) -> str:
        return simple_repr(self, ["focus_concept", "refinement"])



#!/usr/bin/env python
# cardinal_pythonlib/ui.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for user interaction.**

"""

import os

# For back compatibility (these functions used to be in this file):
# noinspection PyUnresolvedReferences
from cardinal_pythonlib.ui_commandline import ask_user, ask_user_password

try:
    import tkinter
    from tkinter import filedialog
except ImportError:
    tkinter = None  # make type checker happy
    filedialog = None  # make type checker happy
    if not os.environ.get("_SPHINX_AUTODOC_IN_PROGRESS", None):
        raise


def get_save_as_filename(defaultfilename: str,
                         defaultextension: str,
                         title: str = "Save As") -> str:
    """
    Provides a GUI "Save As" dialogue (via ``tkinter``) and returns the
    filename.
    """
    root = tkinter.Tk()  # create and get Tk topmost window
    # (don't do this too early; the command prompt loses focus)
    root.withdraw()  # won't need this; this gets rid of a blank Tk window
    root.attributes('-topmost', True)  # makes the tk window topmost
    filename = filedialog.asksaveasfilename(
        initialfile=defaultfilename,
        defaultextension=defaultextension,
        parent=root,
        title=title
    )
    root.attributes('-topmost', False)  # stop the tk window being topmost
    return filename


def get_open_filename(defaultfilename: str,
                      defaultextension: str,
                      title: str = "Open") -> str:
    """
    Provides a GUI "Open" dialogue (via ``tkinter``) and returns the filename.
    """
    root = tkinter.Tk()  # create and get Tk topmost window
    # (don't do this too early; the command prompt loses focus)
    root.withdraw()  # won't need this; this gets rid of a blank Tk window
    root.attributes('-topmost', True)  # makes the tk window topmost
    filename = filedialog.askopenfilename(
        initialfile=defaultfilename,
        defaultextension=defaultextension,
        parent=root,
        title=title
    )
    root.attributes('-topmost', False)  # stop the tk window being topmost
    return filename



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/alembic_func.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for Alembic, the migration tool for SQLAlchemy.**

"""

import os
import re
import subprocess
from typing import Tuple

from alembic.config import Config
from alembic.util.exc import CommandError
from alembic.runtime.migration import MigrationContext
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from sqlalchemy.engine import create_engine

from cardinal_pythonlib.fileops import preserve_cwd
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Constants for Alembic
# =============================================================================
# https://alembic.readthedocs.org/en/latest/naming.html
# http://docs.sqlalchemy.org/en/latest/core/constraints.html#configuring-constraint-naming-conventions  # noqa

ALEMBIC_NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    # "ck": "ck_%(table_name)s_%(constraint_name)s",  # too long?
    # ... https://groups.google.com/forum/#!topic/sqlalchemy/SIT4D8S9dUg
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}

DEFAULT_ALEMBIC_VERSION_TABLE = "alembic_version"


# =============================================================================
# Alembic revision/migration system
# =============================================================================
# https://stackoverflow.com/questions/24622170/using-alembic-api-from-inside-application-code  # noqa

def get_head_revision_from_alembic(
        alembic_config_filename: str,
        alembic_base_dir: str = None,
        version_table: str = DEFAULT_ALEMBIC_VERSION_TABLE) -> str:
    """
    Ask Alembic what its head revision is (i.e. where the Python code would
    like the database to be at).

    Arguments:
        alembic_config_filename: config filename
        alembic_base_dir: directory to start in, so relative paths in the
            config file work.
        version_table: table name for Alembic versions
    """
    if alembic_base_dir is None:
        alembic_base_dir = os.path.dirname(alembic_config_filename)
    os.chdir(alembic_base_dir)  # so the directory in the config file works
    config = Config(alembic_config_filename)
    script = ScriptDirectory.from_config(config)
    with EnvironmentContext(config,
                            script,
                            version_table=version_table):
        return script.get_current_head()


def get_current_revision(
        database_url: str,
        version_table: str = DEFAULT_ALEMBIC_VERSION_TABLE) -> str:
    """
    Ask the database what its current revision is.

    Arguments:
        database_url: SQLAlchemy URL for the database
        version_table: table name for Alembic versions
    """
    engine = create_engine(database_url)
    conn = engine.connect()
    opts = {'version_table': version_table}
    mig_context = MigrationContext.configure(conn, opts=opts)
    return mig_context.get_current_revision()


def get_current_and_head_revision(
        database_url: str,
        alembic_config_filename: str,
        alembic_base_dir: str = None,
        version_table: str = DEFAULT_ALEMBIC_VERSION_TABLE) -> Tuple[str, str]:
    """
    Returns a tuple of ``(current_revision, head_revision)``; see
    :func:`get_current_revision` and :func:`get_head_revision_from_alembic`.

    Arguments:
        database_url: SQLAlchemy URL for the database
        alembic_config_filename: config filename
        alembic_base_dir: directory to start in, so relative paths in the
            config file work.
        version_table: table name for Alembic versions
    """
    # Where we are
    head_revision = get_head_revision_from_alembic(
        alembic_config_filename=alembic_config_filename,
        alembic_base_dir=alembic_base_dir,
        version_table=version_table
    )
    log.debug("Intended database version: {}", head_revision)

    # Where we want to be
    current_revision = get_current_revision(
        database_url=database_url,
        version_table=version_table
    )
    log.debug("Current database version: {}", current_revision)

    # Are we where we want to be?
    return current_revision, head_revision


@preserve_cwd
def upgrade_database(
        alembic_config_filename: str,
        alembic_base_dir: str = None,
        starting_revision: str = None,
        destination_revision: str = "head",
        version_table: str = DEFAULT_ALEMBIC_VERSION_TABLE,
        as_sql: bool = False) -> None:
    """
    Use Alembic to upgrade our database.

    See https://alembic.readthedocs.org/en/latest/api/runtime.html
    but also, in particular, ``site-packages/alembic/command.py``

    Arguments:
        alembic_config_filename:
            config filename

        alembic_base_dir:
            directory to start in, so relative paths in the config file work

        starting_revision:
            revision to start at (typically ``None`` to ask the database)

        destination_revision:
            revision to aim for (typically ``"head"`` to migrate to the latest
            structure)

        version_table: table name for Alembic versions

        as_sql:
            run in "offline" mode: print the migration SQL, rather than
            modifying the database. See
            https://alembic.zzzcomputing.com/en/latest/offline.html

    """

    if alembic_base_dir is None:
        alembic_base_dir = os.path.dirname(alembic_config_filename)
    os.chdir(alembic_base_dir)  # so the directory in the config file works
    config = Config(alembic_config_filename)
    script = ScriptDirectory.from_config(config)

    # noinspection PyUnusedLocal,PyProtectedMember
    def upgrade(rev, context):
        return script._upgrade_revs(destination_revision, rev)

    log.info("Upgrading database to revision {!r} using Alembic",
             destination_revision)

    with EnvironmentContext(config,
                            script,
                            fn=upgrade,
                            as_sql=as_sql,
                            starting_rev=starting_revision,
                            destination_rev=destination_revision,
                            tag=None,
                            version_table=version_table):
        script.run_env()

    log.info("Database upgrade completed")


@preserve_cwd
def downgrade_database(
        alembic_config_filename: str,
        destination_revision: str,
        alembic_base_dir: str = None,
        starting_revision: str = None,
        version_table: str = DEFAULT_ALEMBIC_VERSION_TABLE,
        as_sql: bool = False) -> None:
    """
    Use Alembic to downgrade our database. USE WITH EXTREME CAUTION.
    "revision" is the destination revision.

    See https://alembic.readthedocs.org/en/latest/api/runtime.html
    but also, in particular, ``site-packages/alembic/command.py``

    Arguments:
        alembic_config_filename:
            config filename

        alembic_base_dir:
            directory to start in, so relative paths in the config file work

        starting_revision:
            revision to start at (typically ``None`` to ask the database)

        destination_revision:
            revision to aim for

        version_table: table name for Alembic versions

        as_sql:
            run in "offline" mode: print the migration SQL, rather than
            modifying the database. See
            https://alembic.zzzcomputing.com/en/latest/offline.html

    """

    if alembic_base_dir is None:
        alembic_base_dir = os.path.dirname(alembic_config_filename)
    os.chdir(alembic_base_dir)  # so the directory in the config file works
    config = Config(alembic_config_filename)
    script = ScriptDirectory.from_config(config)

    # noinspection PyUnusedLocal,PyProtectedMember
    def downgrade(rev, context):
        return script._downgrade_revs(destination_revision, rev)

    log.info("Downgrading database to revision {!r} using Alembic",
             destination_revision)

    with EnvironmentContext(config,
                            script,
                            fn=downgrade,
                            as_sql=as_sql,
                            starting_rev=starting_revision,
                            destination_rev=destination_revision,
                            tag=None,
                            version_table=version_table):
        script.run_env()

    log.info("Database downgrade completed")


@preserve_cwd
def create_database_migration_numbered_style(
        alembic_ini_file: str,
        alembic_versions_dir: str,
        message: str,
        n_sequence_chars: int = 4) -> None:
    """
    Create a new Alembic migration script.

    Alembic compares the **state of the database** to the **state of the
    metadata**, and generates a migration that brings the former up to the
    latter. (It does **not** compare the most recent revision to the current
    metadata, so make sure your database is up to date with the most recent
    revision before running this!)

    You **must check** that the autogenerated revisions are sensible.

    How does it know where to look for the database?

        1. This function changes into the directory of the Alembic ``.ini``
           file and calls the external program

           .. code-block:: bash

                alembic -c ALEMBIC_INI_FILE revision --autogenerate -m MESSAGE --rev-id REVISION_ID

        2. The Alembic ``.ini`` file points (via the ``script_location``
           variable) to a directory containing your ``env.py``. Alembic loads
           this script.

        3. That script typically works out the database URL and calls further
           into the Alembic code.

   See https://alembic.zzzcomputing.com/en/latest/autogenerate.html.

    Regarding filenames: the default ``n_sequence_chars`` of 4 is like Django
    and gives files with names like

    .. code-block:: none

        0001_x.py, 0002_y.py, ...

    NOTE THAT TO USE A NON-STANDARD ALEMBIC VERSION TABLE, YOU MUST SPECIFY
    THAT IN YOUR ``env.py`` (see e.g. CamCOPS).

    Args:
        alembic_ini_file: filename of Alembic ``alembic.ini`` file
        alembic_versions_dir: directory in which you keep your Python scripts,
            one per Alembic revision
        message: message to be associated with this revision
        n_sequence_chars: number of numerical sequence characters to use in the
            filename/revision (see above).
    """  # noqa
    file_regex = r"\d{" + str(n_sequence_chars) + r"}_\S*\.py$"

    _, _, existing_version_filenames = next(os.walk(alembic_versions_dir),
                                            (None, None, []))
    existing_version_filenames = [
        x for x in existing_version_filenames if re.match(file_regex, x)]
    log.debug("Existing Alembic version script filenames: {!r}",
              existing_version_filenames)
    current_seq_strs = [x[:n_sequence_chars]
                        for x in existing_version_filenames]
    current_seq_strs.sort()
    if not current_seq_strs:
        current_seq_str = None
        new_seq_no = 1
    else:
        current_seq_str = current_seq_strs[-1]
        new_seq_no = max(int(x) for x in current_seq_strs) + 1
    new_seq_str = str(new_seq_no).zfill(n_sequence_chars)

    log.info(
        """
Generating new revision with Alembic...
    Last revision was: {}
    New revision will be: {}
    [If it fails with "Can't locate revision identified by...", you might need
    to DROP the Alembic version table (by default named 'alembic_version', but
    you may have elected to change that in your env.py.]
        """,
        current_seq_str,
        new_seq_str
    )

    alembic_ini_dir = os.path.dirname(alembic_ini_file)
    os.chdir(alembic_ini_dir)
    cmdargs = ['alembic',
               '-c', alembic_ini_file,
               'revision',
               '--autogenerate',
               '-m', message,
               '--rev-id', new_seq_str]
    log.info("From directory {!r}, calling: {!r}", alembic_ini_dir, cmdargs)
    subprocess.call(cmdargs)


def stamp_allowing_unusual_version_table(
        config: Config,
        revision: str,
        sql: bool = False,
        tag: str = None,
        version_table: str = DEFAULT_ALEMBIC_VERSION_TABLE) -> None:
    """
    Stamps the Alembic version table with the given revision; don't run any
    migrations.

    This function is a clone of ``alembic.command.stamp()``, but allowing
    ``version_table`` to change. See
    https://alembic.zzzcomputing.com/en/latest/api/commands.html#alembic.command.stamp
    """  # noqa

    script = ScriptDirectory.from_config(config)

    starting_rev = None
    if ":" in revision:
        if not sql:
            raise CommandError("Range revision not allowed")
        starting_rev, revision = revision.split(':', 2)

    # noinspection PyUnusedLocal
    def do_stamp(rev: str, context):
        # noinspection PyProtectedMember
        return script._stamp_revs(revision, rev)

    with EnvironmentContext(config,
                            script,
                            fn=do_stamp,
                            as_sql=sql,
                            destination_rev=revision,
                            starting_rev=starting_rev,
                            tag=tag,
                            version_table=version_table):
        script.run_env()
